# Copyright (c) 2022 Western Digital Corporation
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import json
import os
import typing as t
import uuid
from http import HTTPStatus
from urllib.error import HTTPError, URLError
from urllib.parse import urlparse

from ansible.module_utils.common.text.converters import to_native
from ansible.module_utils.urls import open_url

if t.TYPE_CHECKING:
    from ansible.module_utils.basic import AnsibleModule


GET_HEADERS = {"accept": "application/json"}
PUT_HEADERS = {"content-type": "application/json", "accept": "application/json"}
POST_HEADERS = {"content-type": "application/json", "accept": "application/json"}
DELETE_HEADERS = {"accept": "application/json"}

HEALTH_OK = 5


class OcapiUtils:
    def __init__(self, creds, base_uri, proxy_slot_number, timeout, module: AnsibleModule) -> None:
        self.root_uri = base_uri
        self.proxy_slot_number = proxy_slot_number
        self.creds = creds
        self.timeout = timeout
        self.module = module

    def _auth_params(self):
        """
        Return tuple of required authentication params based on the username and password.

        :return: tuple of username, password
        """
        username = self.creds["user"]
        password = self.creds["pswd"]
        force_basic_auth = True
        return username, password, force_basic_auth

    def get_request(self, uri):
        req_headers = dict(GET_HEADERS)
        username, password, basic_auth = self._auth_params()
        try:
            resp = open_url(
                uri,
                method="GET",
                headers=req_headers,
                url_username=username,
                url_password=password,
                force_basic_auth=basic_auth,
                validate_certs=False,
                follow_redirects="all",
                use_proxy=True,
                timeout=self.timeout,
            )
            data = json.loads(resp.read())
            headers = {k.lower(): v for (k, v) in resp.info().items()}
        except HTTPError as e:
            return {"ret": False, "msg": f"HTTP Error {e.code} on GET request to '{uri}'", "status": e.code}
        except URLError as e:
            return {"ret": False, "msg": f"URL Error on GET request to '{uri}': '{e.reason}'"}
        # Almost all errors should be caught above, but just in case
        except Exception as e:
            return {"ret": False, "msg": f"Failed GET request to '{uri}': '{e}'"}
        return {"ret": True, "data": data, "headers": headers}

    def delete_request(self, uri, etag=None):
        req_headers = dict(DELETE_HEADERS)
        if etag is not None:
            req_headers["If-Match"] = etag
        username, password, basic_auth = self._auth_params()
        try:
            resp = open_url(
                uri,
                method="DELETE",
                headers=req_headers,
                url_username=username,
                url_password=password,
                force_basic_auth=basic_auth,
                validate_certs=False,
                follow_redirects="all",
                use_proxy=True,
                timeout=self.timeout,
            )
            if resp.status != HTTPStatus.NO_CONTENT:
                data = json.loads(resp.read())
            else:
                data = ""
            headers = {k.lower(): v for (k, v) in resp.info().items()}
        except HTTPError as e:
            return {"ret": False, "msg": f"HTTP Error {e.code} on DELETE request to '{uri}'", "status": e.code}
        except URLError as e:
            return {"ret": False, "msg": f"URL Error on DELETE request to '{uri}': '{e.reason}'"}
        # Almost all errors should be caught above, but just in case
        except Exception as e:
            return {"ret": False, "msg": f"Failed DELETE request to '{uri}': '{e}'"}
        return {"ret": True, "data": data, "headers": headers}

    def put_request(self, uri, payload, etag=None):
        req_headers = dict(PUT_HEADERS)
        if etag is not None:
            req_headers["If-Match"] = etag
        username, password, basic_auth = self._auth_params()
        try:
            resp = open_url(
                uri,
                data=json.dumps(payload),
                headers=req_headers,
                method="PUT",
                url_username=username,
                url_password=password,
                force_basic_auth=basic_auth,
                validate_certs=False,
                follow_redirects="all",
                use_proxy=True,
                timeout=self.timeout,
            )
            headers = {k.lower(): v for (k, v) in resp.info().items()}
        except HTTPError as e:
            return {"ret": False, "msg": f"HTTP Error {e.code} on PUT request to '{uri}'", "status": e.code}
        except URLError as e:
            return {"ret": False, "msg": f"URL Error on PUT request to '{uri}': '{e.reason}'"}
        # Almost all errors should be caught above, but just in case
        except Exception as e:
            return {"ret": False, "msg": f"Failed PUT request to '{uri}': '{e}'"}
        return {"ret": True, "headers": headers, "resp": resp}

    def post_request(self, uri, payload, content_type="application/json", timeout=None):
        req_headers = dict(POST_HEADERS)
        if content_type != "application/json":
            req_headers["content-type"] = content_type
        username, password, basic_auth = self._auth_params()
        if content_type == "application/json":
            request_data = json.dumps(payload)
        else:
            request_data = payload
        try:
            resp = open_url(
                uri,
                data=request_data,
                headers=req_headers,
                method="POST",
                url_username=username,
                url_password=password,
                force_basic_auth=basic_auth,
                validate_certs=False,
                follow_redirects="all",
                use_proxy=True,
                timeout=self.timeout if timeout is None else timeout,
            )
            headers = {k.lower(): v for (k, v) in resp.info().items()}
        except HTTPError as e:
            return {"ret": False, "msg": f"HTTP Error {e.code} on POST request to '{uri}'", "status": e.code}
        except URLError as e:
            return {"ret": False, "msg": f"URL Error on POST request to '{uri}': '{e.reason}'"}
        # Almost all errors should be caught above, but just in case
        except Exception as e:
            return {"ret": False, "msg": f"Failed POST request to '{uri}': '{e}'"}
        return {"ret": True, "headers": headers, "resp": resp}

    def get_uri_with_slot_number_query_param(self, uri):
        """Return the URI with proxy slot number added as a query param, if there is one.

        If a proxy slot number is provided, to access it, we must append it as a query parameter.
        This method returns the given URI with the slotnumber query param added, if there is one.
        If there is not a proxy slot number, it just returns the URI as it was passed in.
        """
        if self.proxy_slot_number is not None:
            parsed_url = urlparse(uri)
            return parsed_url._replace(query=f"slotnumber={self.proxy_slot_number}").geturl()
        else:
            return uri

    def manage_system_power(self, command):
        """Process a command to manage the system power.

        :param str command: The Ansible command being processed.
        """
        if command == "PowerGracefulRestart":
            resource_uri = self.root_uri
            resource_uri = self.get_uri_with_slot_number_query_param(resource_uri)

            # Get the resource so that we have the Etag
            response = self.get_request(resource_uri)
            if "etag" not in response["headers"]:
                return {"ret": False, "msg": "Etag not found in response."}
            etag = response["headers"]["etag"]
            if response["ret"] is False:
                return response

            # Issue the PUT to do the reboot (unless we are in check mode)
            if self.module.check_mode:
                return {"ret": True, "changed": True, "msg": "Update not performed in check mode."}
            payload = {"Reboot": True}
            response = self.put_request(resource_uri, payload, etag)
            if response["ret"] is False:
                return response
        elif command.startswith("PowerMode"):
            return self.manage_power_mode(command)
        else:
            return {"ret": False, "msg": f"Invalid command: {command}"}

        return {"ret": True}

    def manage_chassis_indicator_led(self, command):
        """Process a command to manage the chassis indicator LED.

        :param string command: The Ansible command being processed.
        """
        return self.manage_indicator_led(command, self.root_uri)

    def manage_indicator_led(self, command, resource_uri=None):
        """Process a command to manage an indicator LED.

        :param string command: The Ansible command being processed.
        :param string resource_uri: URI of the resource whose indicator LED is being managed.
        """
        key = "IndicatorLED"
        if resource_uri is None:
            resource_uri = self.root_uri
        resource_uri = self.get_uri_with_slot_number_query_param(resource_uri)

        payloads = {"IndicatorLedOn": {"ID": 2}, "IndicatorLedOff": {"ID": 4}}

        response = self.get_request(resource_uri)
        if "etag" not in response["headers"]:
            return {"ret": False, "msg": "Etag not found in response."}
        etag = response["headers"]["etag"]
        if response["ret"] is False:
            return response
        data = response["data"]
        if key not in data:
            return {"ret": False, "msg": f"Key {key} not found"}
        if "ID" not in data[key]:
            return {"ret": False, "msg": "IndicatorLED for resource has no ID."}

        if command in payloads.keys():
            # See if the LED is already set as requested.
            current_led_status = data[key]["ID"]
            if current_led_status == payloads[command]["ID"]:
                return {"ret": True, "changed": False}

            # Set the LED (unless we are in check mode)
            if self.module.check_mode:
                return {"ret": True, "changed": True, "msg": "Update not performed in check mode."}
            payload = {"IndicatorLED": payloads[command]}
            response = self.put_request(resource_uri, payload, etag)
            if response["ret"] is False:
                return response
        else:
            return {"ret": False, "msg": "Invalid command"}

        return {"ret": True}

    def manage_power_mode(self, command):
        key = "PowerState"
        resource_uri = self.get_uri_with_slot_number_query_param(self.root_uri)

        payloads = {"PowerModeNormal": 2, "PowerModeLow": 4}

        response = self.get_request(resource_uri)
        if "etag" not in response["headers"]:
            return {"ret": False, "msg": "Etag not found in response."}
        etag = response["headers"]["etag"]
        if response["ret"] is False:
            return response
        data = response["data"]
        if key not in data:
            return {"ret": False, "msg": f"Key {key} not found"}
        if "ID" not in data[key]:
            return {"ret": False, "msg": "PowerState for resource has no ID."}

        if command in payloads.keys():
            # See if the PowerState is already set as requested.
            current_power_state = data[key]["ID"]
            if current_power_state == payloads[command]:
                return {"ret": True, "changed": False}

            # Set the Power State (unless we are in check mode)
            if self.module.check_mode:
                return {"ret": True, "changed": True, "msg": "Update not performed in check mode."}
            payload = {"PowerState": {"ID": payloads[command]}}
            response = self.put_request(resource_uri, payload, etag)
            if response["ret"] is False:
                return response
        else:
            return {"ret": False, "msg": f"Invalid command: {command}"}

        return {"ret": True}

    def prepare_multipart_firmware_upload(self, filename):
        """Prepare a multipart/form-data body for OCAPI firmware upload.

        :arg filename: The name of the file to upload.
        :returns: tuple of (content_type, body) where ``content_type`` is
            the ``multipart/form-data`` ``Content-Type`` header including
            ``boundary`` and ``body`` is the prepared bytestring body

        Prepares the body to include "FirmwareFile" field with the contents of the file.
        Because some OCAPI targets do not support Base-64 encoding for multipart/form-data,
        this method sends the file as binary.
        """
        boundary = str(uuid.uuid4())  # Generate a random boundary
        body = f"--{boundary}\r\n"
        body += f'Content-Disposition: form-data; name="FirmwareFile"; filename="{to_native(os.path.basename(filename))}"\r\n'
        body += "Content-Type: application/octet-stream\r\n\r\n"
        body_bytes = bytearray(body, "utf-8")
        with open(filename, "rb") as f:
            body_bytes += f.read()
        body_bytes += bytearray(f"\r\n--{boundary}--", "utf-8")
        return (f"multipart/form-data; boundary={boundary}", body_bytes)

    def upload_firmware_image(self, update_image_path):
        """Perform Firmware Upload to the OCAPI storage device.

        :param str update_image_path: The path/filename of the firmware image, on the local filesystem.
        """
        if not (os.path.exists(update_image_path) and os.path.isfile(update_image_path)):
            return {"ret": False, "msg": "File does not exist."}
        url = f"{self.root_uri}OperatingSystem"
        url = self.get_uri_with_slot_number_query_param(url)
        content_type, b_form_data = self.prepare_multipart_firmware_upload(update_image_path)

        # Post the firmware (unless we are in check mode)
        if self.module.check_mode:
            return {"ret": True, "changed": True, "msg": "Update not performed in check mode."}
        result = self.post_request(url, b_form_data, content_type=content_type, timeout=300)
        if result["ret"] is False:
            return result
        return {"ret": True}

    def update_firmware_image(self):
        """Perform a Firmware Update on the OCAPI storage device."""
        resource_uri = self.root_uri
        resource_uri = self.get_uri_with_slot_number_query_param(resource_uri)
        # We have to do a GET to obtain the Etag.  It's required on the PUT.
        response = self.get_request(resource_uri)
        if response["ret"] is False:
            return response
        if "etag" not in response["headers"]:
            return {"ret": False, "msg": "Etag not found in response."}
        etag = response["headers"]["etag"]

        # Issue the PUT (unless we are in check mode)
        if self.module.check_mode:
            return {"ret": True, "changed": True, "msg": "Update not performed in check mode."}
        payload = {"FirmwareUpdate": True}
        response = self.put_request(resource_uri, payload, etag)
        if response["ret"] is False:
            return response

        return {"ret": True, "jobUri": response["headers"]["location"]}

    def activate_firmware_image(self):
        """Perform a Firmware Activate on the OCAPI storage device."""
        resource_uri = self.root_uri
        resource_uri = self.get_uri_with_slot_number_query_param(resource_uri)
        # We have to do a GET to obtain the Etag.  It's required on the PUT.
        response = self.get_request(resource_uri)
        if "etag" not in response["headers"]:
            return {"ret": False, "msg": "Etag not found in response."}
        etag = response["headers"]["etag"]
        if response["ret"] is False:
            return response

        # Issue the PUT (unless we are in check mode)
        if self.module.check_mode:
            return {"ret": True, "changed": True, "msg": "Update not performed in check mode."}
        payload = {"FirmwareActivate": True}
        response = self.put_request(resource_uri, payload, etag)
        if response["ret"] is False:
            return response

        return {"ret": True, "jobUri": response["headers"]["location"]}

    def get_job_status(self, job_uri):
        """Get the status of a job.

        :param str job_uri: The URI of the job's status monitor.
        """
        job_uri = self.get_uri_with_slot_number_query_param(job_uri)
        response = self.get_request(job_uri)
        if response["ret"] is False:
            if response.get("status") == HTTPStatus.NOT_FOUND:
                # Job not found -- assume 0%
                return {
                    "ret": True,
                    "percentComplete": 0,
                    "operationStatus": "Not Available",
                    "operationStatusId": 1,
                    "operationHealth": None,
                    "operationHealthId": None,
                    "details": "Job does not exist.",
                    "jobExists": False,
                }
            else:
                return response
        details = response["data"]["Status"].get("Details")
        if isinstance(details, str):
            details = [details]
        health_list = response["data"]["Status"]["Health"]
        return_value = {
            "ret": True,
            "percentComplete": response["data"]["PercentComplete"],
            "operationStatus": response["data"]["Status"]["State"]["Name"],
            "operationStatusId": response["data"]["Status"]["State"]["ID"],
            "operationHealth": health_list[0]["Name"] if len(health_list) > 0 else None,
            "operationHealthId": health_list[0]["ID"] if len(health_list) > 0 else None,
            "details": details,
            "jobExists": True,
        }
        return return_value

    def delete_job(self, job_uri):
        """Delete the OCAPI job referenced by the specified job_uri."""
        job_uri = self.get_uri_with_slot_number_query_param(job_uri)
        # We have to do a GET to obtain the Etag.  It's required on the DELETE.
        response = self.get_request(job_uri)

        if response["ret"] is True:
            if "etag" not in response["headers"]:
                return {"ret": False, "msg": "Etag not found in response."}
            else:
                etag = response["headers"]["etag"]

            if response["data"]["PercentComplete"] != 100:
                return {"ret": False, "changed": False, "msg": "Cannot delete job because it is in progress."}

        if response["ret"] is False:
            if response["status"] == HTTPStatus.NOT_FOUND:
                return {"ret": True, "changed": False, "msg": "Job already deleted."}
            return response
        if self.module.check_mode:
            return {"ret": True, "changed": True, "msg": "Update not performed in check mode."}

        # Do the DELETE (unless we are in check mode)
        response = self.delete_request(job_uri, etag)
        if response["ret"] is False:
            if response["status"] == HTTPStatus.NOT_FOUND:
                return {"ret": True, "changed": False}
            elif response["status"] == HTTPStatus.CONFLICT:
                return {"ret": False, "changed": False, "msg": "Cannot delete job because it is in progress."}
            return response
        return {"ret": True, "changed": True}



# Copyright (c) 2021 Jeff Goldschrafe <jeff@holyhandgrenade.org>
# Based on Ansible local connection plugin by:
# Copyright (c) 2012 Michael DeHaan <michael.dehaan@gmail.com>
# Copyright (c) 2015, 2017 Toshio Kuratomi <tkuratomi@ansible.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
name: nsenter
short_description: execute on host running controller container
version_added: 1.9.0
description:
  - This connection plugin allows Ansible, running in a privileged container, to execute tasks on the container host instead
    of in the container itself.
  - This is useful for running Ansible in a pull model, while still keeping the Ansible control node containerized.
  - It relies on having privileged access to run C(nsenter) in the host's PID namespace, allowing it to enter the namespaces
    of the provided PID (default PID 1, or init/systemd).
author: Jeff Goldschrafe (@jgoldschrafe)
options:
  nsenter_pid:
    description:
      - PID to attach with using nsenter.
      - The default should be fine unless you are attaching as a non-root user.
    type: int
    default: 1
    vars:
      - name: ansible_nsenter_pid
    env:
      - name: ANSIBLE_NSENTER_PID
    ini:
      - section: nsenter_connection
        key: nsenter_pid
notes:
  - The remote user is ignored; this plugin always runs as root.
  - "This plugin requires the Ansible controller container to be launched in the following way: (1) The container image contains
    the C(nsenter) program; (2) The container is launched in privileged mode; (3) The container is launched in the host's
    PID namespace (C(--pid host))."
"""

import fcntl
import os
import pty
import selectors
import shlex
import subprocess
import typing as t

import ansible.constants as C
from ansible.errors import AnsibleError
from ansible.module_utils.common.text.converters import to_bytes, to_text
from ansible.plugins.connection import ConnectionBase
from ansible.utils.display import Display
from ansible.utils.path import unfrackpath

display = Display()


class Connection(ConnectionBase):
    """Connections to a container host using nsenter"""

    transport = "community.docker.nsenter"
    has_pipelining = False

    def __init__(self, *args: t.Any, **kwargs: t.Any) -> None:
        super().__init__(*args, **kwargs)
        self.cwd = None
        self._nsenter_pid = None

    def _connect(self) -> t.Self:
        self._nsenter_pid = self.get_option("nsenter_pid")

        # Because nsenter requires very high privileges, our remote user
        # is always assumed to be root.
        self._play_context.remote_user = "root"

        if not self._connected:
            display.vvv(
                f"ESTABLISH NSENTER CONNECTION FOR USER: {self._play_context.remote_user}",
                host=self._play_context.remote_addr,
            )
            self._connected = True
        return self

    def exec_command(
        self, cmd: str, in_data: bytes | None = None, sudoable: bool = True
    ) -> tuple[int, bytes, bytes]:
        super().exec_command(cmd, in_data=in_data, sudoable=sudoable)  # type: ignore[safe-super]

        display.debug("in nsenter.exec_command()")

        # pylint: disable-next=no-member
        def_executable: str | None = C.DEFAULT_EXECUTABLE  # type: ignore[attr-defined]
        executable = def_executable.split()[0] if def_executable else None

        if not os.path.exists(to_bytes(executable, errors="surrogate_or_strict")):
            raise AnsibleError(
                f"failed to find the executable specified {executable}."
                " Please verify if the executable exists and re-try."
            )

        # Rewrite the provided command to prefix it with nsenter
        nsenter_cmd_parts = [
            "nsenter",
            "--ipc",
            "--mount",
            "--net",
            "--pid",
            "--uts",
            "--preserve-credentials",
            f"--target={self._nsenter_pid}",
            "--",
        ]

        cmd_parts = nsenter_cmd_parts + [cmd]
        cmd_b = to_bytes(" ".join(cmd_parts))

        display.vvv(f"EXEC {to_text(cmd_b)}", host=self._play_context.remote_addr)
        display.debug("opening command with Popen()")

        master = None
        stdin = subprocess.PIPE

        # This plugin does not support pipelining. This diverges from the behavior of
        # the core "local" connection plugin that this one derives from.
        if sudoable and self.become and self.become.expect_prompt():
            # Create a pty if sudoable for privilege escalation that needs it.
            # Falls back to using a standard pipe if this fails, which may
            # cause the command to fail in certain situations where we are escalating
            # privileges or the command otherwise needs a pty.
            try:
                master, stdin = pty.openpty()
            except (IOError, OSError) as e:
                display.debug(f"Unable to open pty: {e}")

        with subprocess.Popen(
            cmd_b,
            shell=True,
            executable=executable,
            cwd=self.cwd,
            stdin=stdin,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        ) as p:
            assert p.stderr is not None
            assert p.stdin is not None
            assert p.stdout is not None
            # if we created a master, we can close the other half of the pty now, otherwise master is stdin
            if master is not None:
                os.close(stdin)

            display.debug("done running command with Popen()")

            if self.become and self.become.expect_prompt() and sudoable:
                fcntl.fcntl(
                    p.stdout,
                    fcntl.F_SETFL,
                    fcntl.fcntl(p.stdout, fcntl.F_GETFL) | os.O_NONBLOCK,
                )
                fcntl.fcntl(
                    p.stderr,
                    fcntl.F_SETFL,
                    fcntl.fcntl(p.stderr, fcntl.F_GETFL) | os.O_NONBLOCK,
                )
                selector = selectors.DefaultSelector()
                selector.register(p.stdout, selectors.EVENT_READ)
                selector.register(p.stderr, selectors.EVENT_READ)

                become_output = b""
                try:
                    while not self.become.check_success(
                        become_output
                    ) and not self.become.check_password_prompt(become_output):
                        events = selector.select(self._play_context.timeout)
                        if not events:
                            stdout, stderr = p.communicate()
                            raise AnsibleError(
                                "timeout waiting for privilege escalation password prompt:\n"
                                + to_text(become_output)
                            )

                        chunks = b""
                        for key, dummy_event in events:
                            if key.fileobj == p.stdout:
                                chunk = p.stdout.read()
                                if chunk:
                                    chunks += chunk
                            elif key.fileobj == p.stderr:
                                chunk = p.stderr.read()
                                if chunk:
                                    chunks += chunk

                        if not chunks:
                            stdout, stderr = p.communicate()
                            raise AnsibleError(
                                "privilege output closed while waiting for password prompt:\n"
                                + to_text(become_output)
                            )
                        become_output += chunks
                finally:
                    selector.close()

                if not self.become.check_success(become_output):
                    become_pass = self.become.get_option(
                        "become_pass", playcontext=self._play_context
                    )
                    if master is None:
                        p.stdin.write(
                            to_bytes(become_pass, errors="surrogate_or_strict") + b"\n"
                        )
                    else:
                        os.write(
                            master,
                            to_bytes(become_pass, errors="surrogate_or_strict") + b"\n",
                        )

                fcntl.fcntl(
                    p.stdout,
                    fcntl.F_SETFL,
                    fcntl.fcntl(p.stdout, fcntl.F_GETFL) & ~os.O_NONBLOCK,
                )
                fcntl.fcntl(
                    p.stderr,
                    fcntl.F_SETFL,
                    fcntl.fcntl(p.stderr, fcntl.F_GETFL) & ~os.O_NONBLOCK,
                )

            display.debug("getting output with communicate()")
            stdout, stderr = p.communicate(in_data)
            display.debug("done communicating")

            # finally, close the other half of the pty, if it was created
            if master:
                os.close(master)

            display.debug("done with nsenter.exec_command()")
            return (p.returncode, stdout, stderr)

    def put_file(self, in_path: str, out_path: str) -> None:
        super().put_file(in_path, out_path)  # type: ignore[safe-super]

        in_path = unfrackpath(in_path, basedir=self.cwd)
        out_path = unfrackpath(out_path, basedir=self.cwd)

        display.vvv(f"PUT {in_path} to {out_path}", host=self._play_context.remote_addr)
        try:
            with open(to_bytes(in_path, errors="surrogate_or_strict"), "rb") as in_file:
                in_data = in_file.read()
            rc, dummy_out, err = self.exec_command(
                cmd=f"tee {shlex.quote(out_path)}", in_data=in_data
            )
            if rc != 0:
                raise AnsibleError(
                    f"failed to transfer file to {out_path}: {to_text(err)}"
                )
        except IOError as e:
            raise AnsibleError(f"failed to transfer file to {out_path}: {e}") from e

    def fetch_file(self, in_path: str, out_path: str) -> None:
        super().fetch_file(in_path, out_path)  # type: ignore[safe-super]

        in_path = unfrackpath(in_path, basedir=self.cwd)
        out_path = unfrackpath(out_path, basedir=self.cwd)

        try:
            rc, out, err = self.exec_command(cmd=f"cat {shlex.quote(in_path)}")
            display.vvv(
                f"FETCH {in_path} TO {out_path}", host=self._play_context.remote_addr
            )
            if rc != 0:
                raise AnsibleError(
                    f"failed to transfer file to {in_path}: {to_text(err)}"
                )
            with open(
                to_bytes(out_path, errors="surrogate_or_strict"), "wb"
            ) as out_file:
                out_file.write(out)
        except IOError as e:
            raise AnsibleError(
                f"failed to transfer file to {to_text(out_path)}: {e}"
            ) from e

    def close(self) -> None:
        """terminate the connection; nothing to do here"""
        self._connected = False



# -*- coding: utf-8 -*-
# Copyright 2022 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function


__metaclass__ = type

#############################################
#                WARNING                    #
#############################################
#
# This file is auto generated by the
# cli_rm_builder.
#
# Manually editing this file is not advised.
#
# To update the argspec make the desired changes
# in the module docstring and re-run
# cli_rm_builder.
#
#############################################

"""
The arg spec for the ios_bgp_address_family module
"""


class Bgp_address_familyArgs(object):  # pylint: disable=R0903
    """The arg spec for the ios_bgp_address_family module"""

    argument_spec = {
        "config": {
            "type": "dict",
            "options": {
                "as_number": {"type": "str"},
                "address_family": {
                    "type": "list",
                    "elements": "dict",
                    "options": {
                        "afi": {
                            "type": "str",
                            "choices": [
                                "ipv4",
                                "ipv6",
                                "l2vpn",
                                "nsap",
                                "rtfilter",
                                "vpnv4",
                                "vpnv6",
                            ],
                        },
                        "safi": {
                            "type": "str",
                            "choices": ["flowspec", "mdt", "multicast", "mvpn", "evpn", "unicast"],
                        },
                        "vrf": {"type": "str"},
                        "aggregate_addresses": {
                            "type": "list",
                            "elements": "dict",
                            "aliases": ["aggregate_address"],
                            "options": {
                                "address": {"type": "str"},
                                "netmask": {"type": "str"},
                                "advertise_map": {"type": "str"},
                                "as_confed_set": {"type": "bool"},
                                "as_set": {"type": "bool"},
                                "attribute_map": {"type": "str"},
                                "summary_only": {"type": "bool"},
                                "suppress_map": {"type": "str"},
                            },
                        },
                        "auto_summary": {"type": "bool"},
                        "maximum_paths": {
                            "type": "dict",
                            "options": {
                                "paths": {"type": "int"},
                                "eibgp": {"type": "int"},
                                "ibgp": {"type": "int"},
                            },
                        },
                        "maximum_secondary_paths": {
                            "type": "dict",
                            "options": {
                                "paths": {"type": "int"},
                                "eibgp": {"type": "int"},
                                "ibgp": {"type": "int"},
                            },
                        },
                        "bgp": {
                            "type": "dict",
                            "options": {
                                "additional_paths": {
                                    "type": "dict",
                                    "options": {
                                        "receive": {"type": "bool"},
                                        "select": {
                                            "type": "dict",
                                            "options": {
                                                "all": {"type": "bool"},
                                                "backup": {"type": "bool"},
                                                "best": {"type": "int"},
                                                "best_external": {"type": "bool"},
                                                "group_best": {"type": "bool"},
                                            },
                                        },
                                        "send": {"type": "bool"},
                                        "install": {"type": "bool"},
                                    },
                                },
                                "aggregate_timer": {"type": "int"},
                                "dampening": {
                                    "type": "dict",
                                    "options": {
                                        "penalty_half_time": {"type": "int"},
                                        "reuse_route_val": {"type": "int"},
                                        "suppress_route_val": {"type": "int"},
                                        "max_suppress": {"type": "int"},
                                        "route_map": {"type": "str"},
                                    },
                                },
                                "dmzlink_bw": {"type": "bool"},
                                "nexthop": {
                                    "type": "dict",
                                    "options": {
                                        "route_map": {"type": "str"},
                                        "trigger": {
                                            "type": "dict",
                                            "options": {
                                                "delay": {"type": "int"},
                                                "enable": {"type": "bool"},
                                            },
                                        },
                                    },
                                },
                                "redistribute_internal": {"type": "bool"},
                                "route_map": {"type": "bool"},
                                "scan_time": {"type": "int"},
                                "slow_peer": {
                                    "type": "list",
                                    "elements": "dict",
                                    "options": {
                                        "detection": {
                                            "type": "dict",
                                            "options": {
                                                "enable": {"type": "bool"},
                                                "threshold": {"type": "int"},
                                            },
                                        },
                                        "split_update_group": {
                                            "type": "dict",
                                            "options": {
                                                "dynamic": {"type": "bool"},
                                                "permanent": {"type": "bool"},
                                            },
                                        },
                                    },
                                },
                                "slow_peer_options": {
                                    "type": "dict",
                                    "options": {
                                        "detection": {
                                            "type": "dict",
                                            "options": {
                                                "enable": {"type": "bool"},
                                                "threshold": {"type": "int"},
                                            },
                                        },
                                        "split_update_group": {
                                            "type": "dict",
                                            "options": {
                                                "dynamic": {"type": "bool"},
                                                "permanent": {"type": "bool"},
                                            },
                                        },
                                    },
                                },
                                "soft_reconfig_backup": {"type": "bool"},
                                "update_group": {"type": "bool"},
                            },
                        },
                        "default": {"type": "bool"},
                        "default_information": {"type": "bool"},
                        "default_metric": {"type": "int"},
                        "distance": {
                            "type": "dict",
                            "options": {
                                "external": {"type": "int"},
                                "internal": {"type": "int"},
                                "local": {"type": "int"},
                            },
                        },
                        "neighbors": {
                            "type": "list",
                            "elements": "dict",
                            "aliases": ["neighbor"],
                            "options": {
                                "neighbor_address": {"type": "str"},
                                "address": {"type": "str"},
                                "tag": {"type": "str"},
                                "ipv6_address": {"type": "str", "aliases": ["ipv6_adddress"]},
                                "activate": {"type": "bool"},
                                "additional_paths": {
                                    "type": "dict",
                                    "options": {
                                        "disable": {"type": "bool"},
                                        "receive": {"type": "bool"},
                                        "send": {"type": "bool"},
                                    },
                                },
                                "advertise": {
                                    "type": "dict",
                                    "options": {
                                        "all": {"type": "bool"},
                                        "best": {"type": "int"},
                                        "group_best": {"type": "bool"},
                                    },
                                },
                                "advertises": {
                                    "type": "dict",
                                    "options": {
                                        "additional_paths": {
                                            "type": "dict",
                                            "options": {
                                                "all": {"type": "bool"},
                                                "best": {"type": "int"},
                                                "group_best": {"type": "bool"},
                                            },
                                        },
                                        "best_external": {"type": "bool"},
                                        "diverse_path": {
                                            "type": "dict",
                                            "options": {
                                                "backup": {"type": "bool"},
                                                "mpath": {"type": "bool"},
                                            },
                                        },
                                    },
                                },
                                "advertise_map": {
                                    "type": "dict",
                                    "options": {
                                        "name": {"type": "str"},
                                        "exist_map": {"type": "str"},
                                        "non_exist_map": {"type": "str"},
                                    },
                                },
                                "advertisement_interval": {"type": "int"},
                                "aigp": {
                                    "type": "dict",
                                    "options": {
                                        "enable": {"type": "str"},
                                        "send": {
                                            "type": "dict",
                                            "options": {
                                                "cost_community": {
                                                    "type": "dict",
                                                    "options": {
                                                        "id": {"type": "int"},
                                                        "poi": {
                                                            "type": "dict",
                                                            "options": {
                                                                "igp_cost": {"type": "bool"},
                                                                "pre_bestpath": {"type": "bool"},
                                                                "transitive": {"type": "bool"},
                                                            },
                                                        },
                                                    },
                                                },
                                                "med": {"type": "bool"},
                                            },
                                        },
                                    },
                                },
                                "allow_policy": {"type": "bool"},
                                "allowas_in": {"type": "int"},
                                "as_override": {
                                    "type": "dict",
                                    "options": {
                                        "set": {"type": "bool"},
                                        "split_horizon": {"type": "bool"},
                                    },
                                },
                                "bmp_activate": {
                                    "type": "dict",
                                    "options": {"all": {"type": "bool"}, "server": {"type": "int"}},
                                },
                                "capability": {
                                    "type": "dict",
                                    "options": {
                                        "both": {"type": "bool"},
                                        "receive": {"type": "bool"},
                                        "send": {"type": "bool"},
                                    },
                                },
                                "cluster_id": {"type": "str"},
                                "default_originate": {
                                    "type": "dict",
                                    "options": {
                                        "set": {"type": "bool"},
                                        "route_map": {"type": "str"},
                                    },
                                },
                                "description": {"type": "str"},
                                "disable_connected_check": {"type": "bool"},
                                "distribute_list": {
                                    "type": "dict",
                                    "options": {
                                        "acl": {"type": "str"},
                                        "in": {"type": "bool"},
                                        "out": {"type": "bool"},
                                    },
                                },
                                "dmzlink_bw": {"type": "bool"},
                                "ebgp_multihop": {
                                    "type": "dict",
                                    "options": {
                                        "enable": {"type": "bool"},
                                        "hop_count": {"type": "int"},
                                    },
                                },
                                "fall_over": {
                                    "type": "dict",
                                    "options": {
                                        "bfd": {
                                            "type": "dict",
                                            "options": {
                                                "set": {"type": "bool"},
                                                "multi_hop": {"type": "bool"},
                                                "single_hop": {"type": "bool"},
                                            },
                                        },
                                        "route_map": {"type": "str"},
                                    },
                                },
                                "filter_list": {
                                    "type": "dict",
                                    "options": {
                                        "as_path_acl": {"type": "int"},
                                        "in": {"type": "bool"},
                                        "out": {"type": "bool"},
                                    },
                                },
                                "ha_mode": {
                                    "type": "dict",
                                    "options": {
                                        "set": {"type": "bool"},
                                        "disable": {"type": "bool"},
                                    },
                                },
                                "inherit": {"type": "str"},
                                "internal_vpn_client": {"type": "bool"},
                                "local_as": {
                                    "type": "dict",
                                    "options": {
                                        "set": {"type": "bool"},
                                        "number": {"type": "str"},
                                        "dual_as": {"type": "bool"},
                                        "no_prepend": {
                                            "type": "dict",
                                            "options": {
                                                "set": {"type": "bool"},
                                                "replace_as": {"type": "bool"},
                                            },
                                        },
                                    },
                                },
                                "log_neighbor_changes": {
                                    "type": "dict",
                                    "options": {
                                        "set": {"type": "bool"},
                                        "disable": {"type": "bool"},
                                    },
                                },
                                "maximum_prefix": {
                                    "type": "dict",
                                    "options": {
                                        "number": {"type": "int"},
                                        "threshold_value": {"type": "int"},
                                        "restart": {"type": "int"},
                                        "warning_only": {"type": "bool"},
                                    },
                                },
                                "nexthop_self": {
                                    "type": "dict",
                                    "options": {"set": {"type": "bool"}, "all": {"type": "bool"}},
                                },
                                "next_hop_unchanged": {
                                    "type": "dict",
                                    "options": {
                                        "set": {"type": "bool"},
                                        "allpaths": {"type": "bool"},
                                    },
                                },
                                "password_options": {
                                    "type": "dict",
                                    "no_log": False,
                                    "options": {
                                        "encryption": {"type": "int"},
                                        "pass_key": {"type": "str", "no_log": True},
                                    },
                                },
                                "path_attribute": {
                                    "type": "dict",
                                    "options": {
                                        "discard": {
                                            "type": "dict",
                                            "options": {
                                                "type": {"type": "int"},
                                                "range": {
                                                    "type": "dict",
                                                    "options": {
                                                        "start": {"type": "int"},
                                                        "end": {"type": "int"},
                                                    },
                                                },
                                                "in": {"type": "bool"},
                                            },
                                        },
                                        "treat_as_withdraw": {
                                            "type": "dict",
                                            "options": {
                                                "type": {"type": "int"},
                                                "range": {
                                                    "type": "dict",
                                                    "options": {
                                                        "start": {"type": "int"},
                                                        "end": {"type": "int"},
                                                    },
                                                },
                                                "in": {"type": "bool"},
                                            },
                                        },
                                    },
                                },
                                "peer_group": {"type": "bool"},
                                "peer_group_name": {"type": "str"},
                                "prefix_list": {
                                    "type": "dict",
                                    "options": {
                                        "name": {"type": "str"},
                                        "in": {"type": "bool"},
                                        "out": {"type": "bool"},
                                    },
                                },
                                "prefix_lists": {
                                    "type": "list",
                                    "elements": "dict",
                                    "options": {
                                        "name": {"type": "str"},
                                        "in": {"type": "bool"},
                                        "out": {"type": "bool"},
                                    },
                                },
                                "remote_as": {"type": "str"},
                                "remove_private_as": {
                                    "type": "dict",
                                    "options": {
                                        "set": {"type": "bool"},
                                        "all": {"type": "bool"},
                                        "replace_as": {"type": "bool"},
                                    },
                                },
                                "route_map": {
                                    "type": "dict",
                                    "options": {
                                        "name": {"type": "str"},
                                        "in": {"type": "bool"},
                                        "out": {"type": "bool"},
                                    },
                                },
                                "route_maps": {
                                    "type": "list",
                                    "elements": "dict",
                                    "options": {
                                        "name": {"type": "str"},
                                        "in": {"type": "bool"},
                                        "out": {"type": "bool"},
                                    },
                                },
                                "route_reflector_client": {"type": "bool"},
                                "route_server_client": {"type": "bool"},
                                "send_community": {
                                    "type": "dict",
                                    "options": {
                                        "set": {"type": "bool"},
                                        "both": {"type": "bool"},
                                        "extended": {"type": "bool"},
                                        "standard": {"type": "bool"},
                                    },
                                },
                                "shutdown": {
                                    "type": "dict",
                                    "options": {
                                        "set": {"type": "bool"},
                                        "graceful": {"type": "int"},
                                    },
                                },
                                "slow_peer": {
                                    "type": "list",
                                    "elements": "dict",
                                    "options": {
                                        "detection": {
                                            "type": "dict",
                                            "options": {
                                                "enable": {"type": "bool"},
                                                "disable": {"type": "bool"},
                                                "threshold": {"type": "int"},
                                            },
                                        },
                                        "split_update_group": {
                                            "type": "dict",
                                            "options": {
                                                "dynamic": {
                                                    "type": "dict",
                                                    "options": {
                                                        "enable": {"type": "bool"},
                                                        "disable": {"type": "bool"},
                                                        "permanent": {"type": "bool"},
                                                    },
                                                },
                                                "static": {"type": "bool"},
                                            },
                                        },
                                    },
                                },
                                "slow_peer_options": {
                                    "type": "dict",
                                    "options": {
                                        "detection": {
                                            "type": "dict",
                                            "options": {
                                                "enable": {"type": "bool"},
                                                "disable": {"type": "bool"},
                                                "threshold": {"type": "int"},
                                            },
                                        },
                                        "split_update_group": {
                                            "type": "dict",
                                            "options": {
                                                "dynamic": {
                                                    "type": "dict",
                                                    "options": {
                                                        "enable": {"type": "bool"},
                                                        "disable": {"type": "bool"},
                                                        "permanent": {"type": "bool"},
                                                    },
                                                },
                                                "static": {"type": "bool"},
                                            },
                                        },
                                    },
                                },
                                "soft_reconfiguration": {"type": "bool"},
                                "soo": {"type": "str"},
                                "timers": {
                                    "type": "dict",
                                    "options": {
                                        "interval": {"type": "int"},
                                        "holdtime": {"type": "int"},
                                        "min_holdtime": {"type": "int"},
                                    },
                                },
                                "transport": {
                                    "type": "dict",
                                    "options": {
                                        "connection_mode": {
                                            "type": "dict",
                                            "options": {
                                                "active": {"type": "bool"},
                                                "passive": {"type": "bool"},
                                            },
                                        },
                                        "multi_session": {"type": "bool"},
                                        "path_mtu_discovery": {
                                            "type": "dict",
                                            "options": {
                                                "set": {"type": "bool"},
                                                "disable": {"type": "bool"},
                                            },
                                        },
                                    },
                                },
                                "ttl_security": {"type": "int"},
                                "unsuppress_map": {"type": "str"},
                                "version": {"type": "int"},
                                "weight": {"type": "int"},
                            },
                        },
                        "networks": {
                            "type": "list",
                            "elements": "dict",
                            "aliases": ["network"],
                            "options": {
                                "address": {"type": "str"},
                                "mask": {"type": "str"},
                                "backdoor": {"type": "bool"},
                                "evpn": {"type": "bool"},
                                "route_map": {"type": "str"},
                            },
                        },
                        "redistribute": {
                            "type": "list",
                            "elements": "dict",
                            "options": {
                                "application": {
                                    "type": "dict",
                                    "options": {
                                        "name": {"type": "str"},
                                        "metric": {"type": "int"},
                                        "route_map": {"type": "str"},
                                    },
                                },
                                "bgp": {
                                    "type": "dict",
                                    "options": {
                                        "as_number": {"type": "str"},
                                        "metric": {"type": "int"},
                                        "route_map": {"type": "str"},
                                    },
                                },
                                "connected": {
                                    "type": "dict",
                                    "options": {
                                        "set": {"type": "bool"},
                                        "metric": {"type": "int"},
                                        "route_map": {"type": "str"},
                                    },
                                },
                                "eigrp": {
                                    "type": "dict",
                                    "options": {
                                        "as_number": {"type": "str"},
                                        "metric": {"type": "int"},
                                        "route_map": {"type": "str"},
                                    },
                                },
                                "isis": {
                                    "type": "dict",
                                    "options": {
                                        "area_tag": {"type": "str"},
                                        "clns": {"type": "bool"},
                                        "ip": {"type": "bool"},
                                        "metric": {"type": "int"},
                                        "route_map": {"type": "str"},
                                    },
                                },
                                "iso_igrp": {
                                    "type": "dict",
                                    "options": {
                                        "area_tag": {"type": "str"},
                                        "route_map": {"type": "str"},
                                    },
                                },
                                "lisp": {
                                    "type": "dict",
                                    "options": {
                                        "set": {"type": "bool"},
                                        "metric": {"type": "int"},
                                        "route_map": {"type": "str"},
                                    },
                                },
                                "mobile": {
                                    "type": "dict",
                                    "options": {
                                        "set": {"type": "bool"},
                                        "metric": {"type": "int"},
                                        "route_map": {"type": "str"},
                                    },
                                },
                                "odr": {
                                    "type": "dict",
                                    "options": {
                                        "set": {"type": "bool"},
                                        "metric": {"type": "int"},
                                        "route_map": {"type": "str"},
                                    },
                                },
                                "ospf": {
                                    "type": "dict",
                                    "options": {
                                        "process_id": {"type": "int"},
                                        "include_connected": {"type": "bool"},
                                        "match": {
                                            "type": "dict",
                                            "options": {
                                                "external": {"type": "bool"},
                                                "internal": {"type": "bool"},
                                                "externals": {
                                                    "type": "dict",
                                                    "options": {
                                                        "type_1": {"type": "bool"},
                                                        "type_2": {"type": "bool"},
                                                    },
                                                },
                                                "nssa_external": {"type": "bool"},
                                                "nssa_externals": {
                                                    "type": "dict",
                                                    "options": {
                                                        "type_1": {"type": "bool"},
                                                        "type_2": {"type": "bool"},
                                                    },
                                                },
                                                "type_1": {"type": "bool"},
                                                "type_2": {"type": "bool"},
                                            },
                                        },
                                        "metric": {"type": "int"},
                                        "route_map": {"type": "str"},
                                        "vrf": {"type": "str"},
                                    },
                                },
                                "ospfv3": {
                                    "type": "dict",
                                    "options": {
                                        "process_id": {"type": "int"},
                                        "match": {
                                            "type": "dict",
                                            "options": {
                                                "external": {"type": "bool"},
                                                "internal": {"type": "bool"},
                                                "externals": {
                                                    "type": "dict",
                                                    "options": {
                                                        "type_1": {"type": "bool"},
                                                        "type_2": {"type": "bool"},
                                                    },
                                                },
                                                "nssa_external": {"type": "bool"},
                                                "nssa_externals": {
                                                    "type": "dict",
                                                    "options": {
                                                        "type_1": {"type": "bool"},
                                                        "type_2": {"type": "bool"},
                                                    },
                                                },
                                                "type_1": {"type": "bool"},
                                                "type_2": {"type": "bool"},
                                            },
                                        },
                                        "metric": {"type": "int"},
                                        "route_map": {"type": "str"},
                                    },
                                },
                                "rip": {
                                    "type": "dict",
                                    "options": {
                                        "set": {"type": "bool"},
                                        "metric": {"type": "int"},
                                        "route_map": {"type": "str"},
                                    },
                                },
                                "static": {
                                    "type": "dict",
                                    "options": {
                                        "set": {"type": "bool"},
                                        "clns": {"type": "bool"},
                                        "ip": {"type": "bool"},
                                        "metric": {"type": "int"},
                                        "route_map": {"type": "str"},
                                    },
                                },
                                "vrf": {
                                    "type": "dict",
                                    "options": {
                                        "name": {"type": "str"},
                                        "global": {"type": "bool"},
                                    },
                                },
                            },
                        },
                        "snmp": {
                            "type": "dict",
                            "options": {
                                "context": {
                                    "type": "dict",
                                    "options": {
                                        "name": {"type": "str"},
                                        "community": {
                                            "type": "dict",
                                            "options": {
                                                "snmp_community": {"type": "str"},
                                                "acl": {"type": "str"},
                                                "ipv6": {"type": "str"},
                                                "ro": {"type": "bool"},
                                                "rw": {"type": "bool"},
                                            },
                                        },
                                        "user": {
                                            "type": "dict",
                                            "options": {
                                                "name": {"type": "str"},
                                                "access": {
                                                    "type": "dict",
                                                    "options": {
                                                        "acl": {"type": "str"},
                                                        "ipv6": {"type": "str"},
                                                    },
                                                },
                                                "auth": {
                                                    "type": "dict",
                                                    "options": {
                                                        "md5": {"type": "str"},
                                                        "sha": {"type": "str"},
                                                    },
                                                },
                                                "priv": {
                                                    "type": "dict",
                                                    "options": {
                                                        "des": {"type": "str"},
                                                        "des56": {"type": "str"},
                                                        "aes128": {"type": "str"},
                                                        "aes192": {"type": "str"},
                                                        "aes256": {"type": "str"},
                                                    },
                                                },
                                                "credential": {"type": "bool"},
                                                "encrypted": {"type": "bool"},
                                            },
                                        },
                                    },
                                },
                            },
                        },
                        "table_map": {
                            "type": "dict",
                            "options": {"name": {"type": "str"}, "filter": {"type": "bool"}},
                        },
                        "advertise": {
                            "type": "dict",
                            "options": {
                                "afi": {
                                    "type": "str",
                                    "choices": ["l2vpn"],
                                },
                                "safi": {
                                    "type": "str",
                                    "choices": ["evpn"],
                                },
                            },
                        },
                    },
                },
            },
        },
        "running_config": {"type": "str"},
        "state": {
            "choices": [
                "merged",
                "replaced",
                "overridden",
                "deleted",
                "gathered",
                "rendered",
                "parsed",
            ],
            "default": "merged",
            "type": "str",
        },
    }  # pylint: disable=C0301



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2019, Dag Wieers (@dagwieers) <dag@wieers.com>
# Copyright: (c) 2025, Gaspard Micol (@gmicol) <gmicol@cisco.com>
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {"metadata_version": "1.1", "status": ["preview"], "supported_by": "community"}

DOCUMENTATION = r"""
---
module: mso_schema_site_anp_epg
short_description: Manage site-local Endpoint Groups (EPGs) in schema template
description:
- Manage site-local EPGs in schema template on Cisco ACI Multi-Site.
author:
- Dag Wieers (@dagwieers)
- Gaspard Micol (@gmicol)
options:
  schema:
    description:
    - The name of the schema.
    type: str
    required: true
  site:
    description:
    - The name of the site.
    type: str
    required: true
  template:
    description:
    - The name of the template.
    type: str
    required: true
  anp:
    description:
    - The name of the ANP.
    type: str
    required: true
  epg:
    description:
    - The name of the EPG to manage.
    type: str
    aliases: [ name ]
  private_link_label:
    description:
    - The private link label used to represent this subnet.
    - This parameter is available for MSO version greater than 3.3.
    type: str
  admin_state:
    description:
    - The EPG admin state.
    - Defaults to C(admin_up) when unset during creation.
    type: str
    choices: [ admin_up, admin_shut ]
  state:
    description:
    - Use C(present) or C(absent) for adding or removing.
    - Use C(query) for listing an object or multiple objects.
    type: str
    choices: [ absent, present, query ]
    default: present
seealso:
- module: cisco.mso.mso_schema_site_anp
- module: cisco.mso.mso_schema_site_anp_epg_subnet
- module: cisco.mso.mso_schema_template_anp_epg
extends_documentation_fragment: cisco.mso.modules
"""

EXAMPLES = r"""
- name: Add a new site EPG
  cisco.mso.mso_schema_site_anp_epg:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema1
    site: Site1
    template: Template1
    anp: ANP1
    epg: EPG1
    state: present

- name: Remove a site EPG
  cisco.mso.mso_schema_site_anp_epg:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema1
    site: Site1
    template: Template1
    anp: ANP1
    epg: EPG1
    state: absent

- name: Query a specific site EPGs
  cisco.mso.mso_schema_site_anp_epg:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema1
    site: Site1
    template: Template1
    anp: ANP1
    epg: EPG1
    state: query
  register: query_result

- name: Query all site EPGs
  cisco.mso.mso_schema_site_anp_epg:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema1
    site: Site1
    template: Template1
    anp: ANP1
    state: query
  register: query_result
"""

RETURN = r"""
"""

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.mso.plugins.module_utils.mso import MSOModule, mso_argument_spec


def main():
    argument_spec = mso_argument_spec()
    argument_spec.update(
        schema=dict(type="str", required=True),
        site=dict(type="str", required=True),
        template=dict(type="str", required=True),
        anp=dict(type="str", required=True),
        epg=dict(type="str", aliases=["name"]),  # This parameter is not required for querying all objects
        private_link_label=dict(type="str"),
        admin_state=dict(type="str", choices=["admin_up", "admin_shut"]),
        state=dict(type="str", default="present", choices=["absent", "present", "query"]),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_if=[
            ["state", "absent", ["epg"]],
            ["state", "present", ["epg"]],
        ],
    )

    schema = module.params.get("schema")
    site = module.params.get("site")
    template = module.params.get("template").replace(" ", "")
    anp = module.params.get("anp")
    epg = module.params.get("epg")
    private_link_label = module.params.get("private_link_label")
    admin_state = module.params.get("admin_state")
    state = module.params.get("state")

    mso = MSOModule(module)

    # Get schema objects
    schema_id, schema_path, schema_obj = mso.query_schema(schema)

    # Get template
    templates = [t.get("name") for t in schema_obj.get("templates")]
    if template not in templates:
        mso.fail_json(msg="Provided template '{0}' does not exist. Existing templates: {1}".format(template, ", ".join(templates)))
    template_idx = templates.index(template)

    # Get site
    site_id = mso.lookup_site(site)

    # Get site_idx
    if not schema_obj.get("sites"):
        mso.fail_json(msg="No site associated with template '{0}'. Associate the site with the template using mso_schema_site.".format(template))
    sites = [(s.get("siteId"), s.get("templateName")) for s in schema_obj.get("sites")]
    sites_list = [s.get("siteId") + "/" + s.get("templateName") for s in schema_obj.get("sites")]
    if (site_id, template) not in sites:
        mso.fail_json(
            msg="Provided site/siteId/template '{0}/{1}/{2}' does not exist. "
            "Existing siteIds/templates: {3}".format(site, site_id, template, ", ".join(sites_list))
        )

    # Schema-access uses indexes
    site_idx = sites.index((site_id, template))
    # Path-based access uses site_id-template
    site_template = "{0}-{1}".format(site_id, template)

    payload = {}
    ops = []
    op_path = ""
    epg_path = None

    # Get ANP
    anp_ref = mso.anp_ref(schema_id=schema_id, template=template, anp=anp)
    anps = [a.get("anpRef") for a in schema_obj["sites"][site_idx]["anps"]]
    anps_in_temp = [a.get("name") for a in schema_obj["templates"][template_idx]["anps"]]
    if anp not in anps_in_temp:
        mso.fail_json(msg="Provided anp '{0}' does not exist. Existing anps: {1}".format(anp, ", ".join(anps_in_temp)))
    else:
        # Get anp index at template level
        template_anp_idx = anps_in_temp.index(anp)

    # If anp not at site level but exists at template level
    if anp_ref not in anps:
        op_path = "/sites/{0}/anps".format(site_template)
        payload = dict(
            anpRef=dict(
                schemaId=schema_id,
                templateName=template,
                anpName=anp,
            ),
        )
    else:
        # Get anp index at site level
        anp_idx = anps.index(anp_ref)

    if epg is not None:
        # Get EPG
        epg_ref = mso.epg_ref(schema_id=schema_id, template=template, anp=anp, epg=epg)
        new_epg = dict(
            epgRef=dict(
                schemaId=schema_id,
                templateName=template,
                anpName=anp,
                epgName=epg,
            )
        )

        # If anp exists at site level
        if "anpRef" not in payload:
            epgs = [e.get("epgRef") for e in schema_obj["sites"][site_idx]["anps"][anp_idx]["epgs"]]

        # If anp already at site level AND if epg not at site level (or) anp not at site level?
        if ("anpRef" not in payload and epg_ref not in epgs) or "anpRef" in payload:
            epgs_in_temp = [e.get("name") for e in schema_obj["templates"][template_idx]["anps"][template_anp_idx]["epgs"]]

            # If EPG not at template level - Fail
            if epg not in epgs_in_temp:
                mso.fail_json(msg="Provided EPG '{0}' does not exist. Existing EPGs: {1}".format(epg, ", ".join(epgs_in_temp)))

            # EPG at template level but not at site level. Create payload at site level for EPG
            else:
                # If anp not in payload then, anp already exists at site level. New payload will only have new EPG payload
                if "anpRef" not in payload:
                    op_path = "/sites/{0}/anps/{1}/epgs".format(site_template, anp)
                    payload = new_epg
                else:
                    # If anp in payload, anp exists at site level. Update payload with EPG payload
                    payload["epgs"] = [new_epg]

        # Get index of EPG at site level
        else:
            epg_idx = epgs.index(epg_ref)
            epg_path = "/sites/{0}/anps/{1}/epgs/{2}".format(site_template, anp, epg)
            mso.existing = schema_obj.get("sites")[site_idx]["anps"][anp_idx]["epgs"][epg_idx]
            payload = new_epg

    ops = []

    if state == "query":
        if anp_ref not in anps:
            mso.fail_json(msg="Provided anp '{0}' does not exist at site level.".format(anp))
        if epg is None:
            mso.existing = schema_obj.get("sites")[site_idx]["anps"][anp_idx]["epgs"]
        elif not mso.existing:
            mso.fail_json(msg="EPG '{epg}' not found".format(epg=epg))
        mso.exit_json()

    # Workaround due to inconsistency in attributes REQUEST/RESPONSE API
    # FIX for MSO Error 400: Bad Request: (0)(0)(0)(0)/deploymentImmediacy error.path.missing
    mso.replace_keys_in_dict("deployImmediacy", "deploymentImmediacy")
    if mso.existing.get("epgRef"):
        epg_ref = mso.dict_from_ref(mso.existing.get("epgRef"))
        mso.existing["epgRef"] = epg_ref

    mso.previous = mso.existing

    if state == "absent":
        if mso.existing and epg_path:
            mso.sent = mso.existing = {}
            ops.append(dict(op="remove", path=epg_path))

    elif state == "present":
        if private_link_label is not None:
            payload["privateLinkLabel"] = dict(name=private_link_label)
        if admin_state is not None:
            payload["shutdown"] = True if admin_state == "admin_shut" else False

        mso.sanitize(payload, collate=True)

        if mso.existing and epg_path:
            ops.append(dict(op="replace", path=epg_path, value=mso.sent))
        else:
            ops.append(dict(op="add", path=op_path + "/-", value=mso.sent))

        mso.existing = mso.proposed

    if not module.check_mode and mso.existing != mso.previous:
        mso.request(schema_path, method="PATCH", data=ops)

    mso.exit_json()


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or
# https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
module: devices_cellular_gateway_lan
short_description: Resource module for devices _cellular _gateway _lan
description:
  - Manage operation update of the resource devices _cellular _gateway _lan.
  - Update the LAN Settings for a single MG.
version_added: '1.0.0'
extends_documentation_fragment:
  - cisco.meraki.module
author: Francisco Munoz (@fmunoz)
options:
  fixedIpAssignments:
    description: List of all fixed IP assignments for a single MG.
    elements: dict
    suboptions:
      ip:
        description: The IP address you want to assign to a specific server or device.
        type: str
      mac:
        description: The MAC address of the server or device that hosts the internal
          resource that you wish to receive the specified IP address.
        type: str
      name:
        description: A descriptive name of the assignment.
        type: str
    type: list
  reservedIpRanges:
    description: List of all reserved IP ranges for a single MG.
    elements: dict
    suboptions:
      comment:
        description: Comment explaining the reserved IP range.
        type: str
      end:
        description: Ending IP included in the reserved range of IPs.
        type: str
      start:
        description: Starting IP included in the reserved range of IPs.
        type: str
    type: list
  serial:
    description: Serial path parameter.
    type: str
requirements:
  - meraki >= 2.4.9
  - python >= 3.5
seealso:
  - name: Cisco Meraki documentation for cellularGateway updateDeviceCellularGatewayLan
    description: Complete reference of the updateDeviceCellularGatewayLan API.
    link: https://developer.cisco.com/meraki/api-v1/#!update-device-cellular-gateway-lan
notes:
  - SDK Method used are
    cellular_gateway.CellularGateway.update_device_cellular_gateway_lan,
  - Paths used are
    put /devices/{serial}/cellularGateway/lan,
"""

EXAMPLES = r"""
- name: Update all
  cisco.meraki.devices_cellular_gateway_lan:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    state: present
    fixedIpAssignments:
      - ip: 192.168.0.10
        mac: 0b:00:00:00:00:ac
        name: server 1
    reservedIpRanges:
      - comment: A reserved IP range
        end: 192.168.1.1
        start: 192.168.1.0
    serial: string
"""
RETURN = r"""
meraki_response:
  description: A dictionary or list with the response returned by the Cisco Meraki Python SDK
  returned: always
  type: dict
  sample: >
    {
      "deviceName": "string",
      "deviceLanIp": "string",
      "deviceSubnet": "string",
      "fixedIpAssignments": [
        {
          "name": "string",
          "ip": "string",
          "mac": "string"
        }
      ],
      "reservedIpRanges": [
        {
          "start": "string",
          "end": "string",
          "comment": "string"
        }
      ]
    }
"""



# Copyright (c) 2018, Stefan Heitmueller <stefan.heitmueller@gmx.com>
# Copyright (c) 2018 Ansible Project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
name: gitlab_runners
author:
  - Stefan Heitmüller (@morph027) <stefan.heitmueller@gmx.com>
short_description: Ansible dynamic inventory plugin for GitLab runners
requirements:
  - python-gitlab > 1.8.0
extends_documentation_fragment:
  - ansible.builtin.constructed
description:
  - Reads inventories from the GitLab API.
  - Uses a YAML configuration file gitlab_runners.[yml|yaml].
options:
  plugin:
    description: The name of this plugin, it should always be set to V(gitlab_runners) for this plugin to recognize it as its own.
    type: str
    required: true
    choices:
      - gitlab_runners
      - community.general.gitlab_runners
  server_url:
    description: The URL of the GitLab server, with protocol (i.e. http or https).
    env:
      - name: GITLAB_SERVER_URL
        version_added: 1.0.0
    type: str
    required: true
  api_token:
    description: GitLab token for logging in.
    env:
      - name: GITLAB_API_TOKEN
        version_added: 1.0.0
    type: str
    aliases:
      - private_token
      - access_token
  filter:
    description: Filter runners from GitLab API.
    env:
      - name: GITLAB_FILTER
        version_added: 1.0.0
    type: str
    choices: ['active', 'paused', 'online', 'specific', 'shared']
  verbose_output:
    description: Toggle to (not) include all available nodes metadata.
    type: bool
    default: true
"""

EXAMPLES = r"""
---
# gitlab_runners.yml
plugin: community.general.gitlab_runners
host: https://gitlab.com

---
# Example using constructed features to create groups and set ansible_host
plugin: community.general.gitlab_runners
host: https://gitlab.com
strict: false
keyed_groups:
  # add e.g. amd64 hosts to an arch_amd64 group
  - prefix: arch
    key: 'architecture'
  # add e.g. linux hosts to an os_linux group
  - prefix: os
    key: 'platform'
  # create a group per runner tag
  # e.g. a runner tagged w/ "production" ends up in group "label_production"
  # hint: labels containing special characters will be converted to safe names
  - key: 'tag_list'
    prefix: tag
"""

from ansible.errors import AnsibleError, AnsibleParserError
from ansible.plugins.inventory import BaseInventoryPlugin, Constructable

from ansible_collections.community.general.plugins.plugin_utils._unsafe import make_unsafe

try:
    import gitlab

    HAS_GITLAB = True
except ImportError:
    HAS_GITLAB = False


class InventoryModule(BaseInventoryPlugin, Constructable):
    """Host inventory parser for ansible using GitLab API as source."""

    NAME = "community.general.gitlab_runners"

    def _populate(self):
        gl = gitlab.Gitlab(self.get_option("server_url"), private_token=self.get_option("api_token"))
        self.inventory.add_group("gitlab_runners")
        try:
            if self.get_option("filter"):
                runners = gl.runners.all(scope=self.get_option("filter"))
            else:
                runners = gl.runners.all()
            for runner in runners:
                host = make_unsafe(str(runner["id"]))
                ip_address = runner["ip_address"]
                host_attrs = make_unsafe(vars(gl.runners.get(runner["id"]))["_attrs"])
                self.inventory.add_host(host, group="gitlab_runners")
                self.inventory.set_variable(host, "ansible_host", make_unsafe(ip_address))
                if self.get_option("verbose_output", True):
                    self.inventory.set_variable(host, "gitlab_runner_attributes", host_attrs)

                # Use constructed if applicable
                strict = self.get_option("strict")
                # Composed variables
                self._set_composite_vars(self.get_option("compose"), host_attrs, host, strict=strict)
                # Complex groups based on jinja2 conditionals, hosts that meet the conditional are added to group
                self._add_host_to_composed_groups(self.get_option("groups"), host_attrs, host, strict=strict)
                # Create groups based on variable values and add the corresponding hosts to it
                self._add_host_to_keyed_groups(self.get_option("keyed_groups"), host_attrs, host, strict=strict)
        except Exception as e:
            raise AnsibleParserError(
                f"Unable to fetch hosts from GitLab API, this was the original exception: {e}"
            ) from e

    def verify_file(self, path):
        """Return the possibly of a file being consumable by this plugin."""
        return super().verify_file(path) and path.endswith(("gitlab_runners.yaml", "gitlab_runners.yml"))

    def parse(self, inventory, loader, path, cache=True):
        if not HAS_GITLAB:
            raise AnsibleError(
                "The GitLab runners dynamic inventory plugin requires python-gitlab: https://python-gitlab.readthedocs.io/en/stable/"
            )
        super().parse(inventory, loader, path, cache)
        self._read_config_data(path)
        self._populate()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or
# https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
module: organizations_wireless_air_marshal_settings_by_network_info
short_description: Information module for organizations _wireless _air _marshal _settings
  _by _network
description:
  - Get all organizations _wireless _air _marshal _settings _by _network.
  - Returns the current Air Marshal settings for this network.
version_added: '1.0.0'
extends_documentation_fragment:
  - cisco.meraki.module_info
  - cisco.meraki.module_info_pagination
author: Francisco Munoz (@fmunoz)
options:
  headers:
    description: Additional headers.
    type: dict
  organizationId:
    description:
      - OrganizationId path parameter. Organization ID.
    type: str
  networkIds:
    description:
      - NetworkIds query parameter. The network IDs to include in the result set.
    elements: str
    type: list
  perPage:
    description:
      - PerPage query parameter. The number of entries per page returned. Acceptable
        range is 3 - 1000. Default is 1000.
    type: int
  startingAfter:
    description:
      - >
        StartingAfter query parameter. A token used by the server to indicate the
        start of the page. Often this is a timestamp or an ID but it is not limited
        to those. This parameter should not be defined by client applications. The
        link for the first, last, prev, or next page in the HTTP Link header should
        define it.
    type: str
  endingBefore:
    description:
      - >
        EndingBefore query parameter. A token used by the server to indicate the end
        of the page. Often this is a timestamp or an ID but it is not limited to those.
        This parameter should not be defined by client applications. The link for
        the first, last, prev, or next page in the HTTP Link header should define
        it.
    type: str
requirements:
  - meraki >= 2.4.9
  - python >= 3.5
seealso:
  - name: Cisco Meraki documentation for wireless getOrganizationWirelessAirMarshalSettingsByNetwork
    description: Complete reference of the getOrganizationWirelessAirMarshalSettingsByNetwork
      API.
    link: https://developer.cisco.com/meraki/api-v1/#!get-organization-wireless-air-marshal-settings-by-network
notes:
  - SDK Method used are
    wireless.Wireless.get_organization_wireless_air_marshal_settings_by_network,
  - Paths used are
    get /organizations/{organizationId}/wireless/airMarshal/settings/byNetwork,
"""

EXAMPLES = r"""
- name: Get all organizations _wireless _air _marshal _settings _by _network
  cisco.meraki.organizations_wireless_air_marshal_settings_by_network_info:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    networkIds: []
    perPage: 0
    startingAfter: string
    endingBefore: string
    organizationId: string
    total_pages: -1
    direction: next
  register: result
"""
RETURN = r"""
meraki_response:
  description: A dictionary or list with the response returned by the Cisco Meraki Python SDK
  returned: always
  type: dict
  sample: >
    {
      "items": [
        {
          "networkId": "string",
          "defaultPolicy": "string"
        }
      ],
      "meta": {
        "counts": {
          "items": {
            "total": 0,
            "remaining": 0
          }
        }
      }
    }
"""



#
# -*- coding: utf-8 -*-
# Copyright 2025 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
#

from __future__ import absolute_import, division, print_function


__metaclass__ = type

"""
The nxos_l2_interfaces config file.
It is in this file where the current configuration (as dict)
is compared to the provided configuration (as dict) and the command set
necessary to bring the current configuration to its desired end-state is
created.
"""

from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.rm_base.resource_module import (
    ResourceModule,
)
from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.utils import (
    dict_merge,
)

from ansible_collections.cisco.nxos.plugins.module_utils.network.nxos.facts.facts import (
    Facts,
)
from ansible_collections.cisco.nxos.plugins.module_utils.network.nxos.rm_templates.l2_interfaces import (
    L2_interfacesTemplate,
)
from ansible_collections.cisco.nxos.plugins.module_utils.network.nxos.utils.utils import (
    generate_switchport_trunk,
    normalize_interface,
    vlan_list_to_range,
    vlan_range_to_dict,
)


class L2_interfaces(ResourceModule):
    """
    The nxos_l2_interfaces config class
    """

    def __init__(self, module):
        super(L2_interfaces, self).__init__(
            empty_fact_val={},
            facts_module=Facts(module),
            module=module,
            resource="l2_interfaces",
            tmplt=L2_interfacesTemplate(),
        )
        self.parsers = [
            "mode",
            "access.vlan",
            "trunk.native_vlan",
            "trunk.allowed_vlans_none",
            "beacon",
            "link_flap.error_disable",
            "cdp_enable",
            "no_cdp_enable",
        ]

    def execute_module(self):
        """Execute the module

        :rtype: A dictionary
        :returns: The result from module execution
        """
        if self.state not in ["parsed", "gathered"]:
            self.generate_commands()
            self.run_commands()
        return self.result

    def generate_commands(self):
        """Generate configuration commands to send based on
        want, have and desired state.
        """
        wantd = {entry["name"].lower(): entry for entry in self.want}
        haved = {entry["name"].lower(): entry for entry in self.have}

        for each in wantd, haved:
            self.process_list_attrs(each)

        # if state is merged, merge want onto have and then compare
        if self.state == "merged":
            wantd = dict_merge(haved, wantd)

        # if state is deleted, empty out wantd and set haved to wantd
        if self.state == "deleted":
            haved = {k: v for k, v in haved.items() if k in wantd or not wantd}
            wantd = {}

        # remove superfluous config for overridden and deleted
        if self.state in ["overridden", "deleted"]:
            for k, have in haved.items():
                if k not in wantd:
                    self._compare(want={}, have=have)

        for k, want in wantd.items():
            self._compare(want=want, have=haved.pop(k, {}))

    def _compare(self, want, have):
        """Leverages the base class `compare()` method and
        populates the list of commands to be run by comparing
        the `want` and `have` data with the `parsers` defined
        for the L2_interfaces network resource.
        """
        begin = len(self.commands)
        want_without_name = want.copy()
        want_without_name.pop("name", None)
        pre_pop_want = bool(want_without_name)
        want_cdp = want.pop("cdp_enable", None)
        have_cdp = have.pop("cdp_enable", None)
        self.handle_cdp(want_cdp, have_cdp, "cdp_enable", pre_pop_want)
        self.compare(parsers=self.parsers, want=want, have=have)
        self._compare_lists(want, have)
        if len(self.commands) != begin:
            self.commands.insert(begin, self._tmplt.render(want or have, "name", False))

    def _compare_lists(self, want, have):
        """Compare list attributes for trunk allowed vlans.

        want_set: VLANs that should be present on the appliance
        have_set: VLANs that are currently on the appliance
        """
        want_set = set(want.get("trunk", {}).get("allowed_vlans", {}))
        have_set = set(have.get("trunk", {}).get("allowed_vlans", {}))

        if self.state == "deleted" and have_set:
            # For deleted state, remove all trunk allowed vlans if any exist
            if have_set:
                self.commands.append("no switchport trunk allowed vlan")
            return

        if self.state in ("merged", "rendered"):
            # Merged/rendered: only ADD vlans, never remove
            # Add vlans that are in want but not in have
            vlans_to_add = want_set - have_set
            if vlans_to_add:
                self.commands.extend(
                    generate_switchport_trunk(
                        "allowed",
                        True,
                        vlan_list_to_range(sorted(vlans_to_add, key=int)),
                    ),
                )
            return

        if self.state in ("replaced", "overridden"):
            # Replaced/overridden: make device match want_set exactly
            # Remove vlans that are in have but not in want
            vlans_to_remove = have_set - want_set
            # Add vlans that are in want but not in have
            vlans_to_add = want_set - have_set

            if not want_set and have_set:
                # Want is empty but have has vlans - remove all
                self.commands.append("no switchport trunk allowed vlan")
            elif vlans_to_remove:
                # Remove excess vlans first
                self.commands.append(
                    f"switchport trunk allowed vlan remove {vlan_list_to_range(sorted(vlans_to_remove, key=int))}",
                )

            if vlans_to_add:
                # Add missing vlans
                self.commands.extend(
                    generate_switchport_trunk(
                        "allowed",
                        True,
                        vlan_list_to_range(sorted(vlans_to_add, key=int)),
                    ),
                )

    def process_list_attrs(self, param):
        if param:
            for _k, val in param.items():
                val["name"] = normalize_interface(val["name"])
                if val.get("trunk"):
                    for vlan in ["allowed_vlans"]:
                        vlanList = val.get("trunk").get(vlan, [])
                        if vlanList and vlanList != "none":
                            val["trunk"][vlan] = vlan_range_to_dict(val.get("trunk").get(vlan))

    def handle_cdp(self, want_cdp, have_cdp, parser, want):
        if want_cdp is None and have_cdp is None:
            if self.state == "replaced" or (self.state == "overridden" and want):
                self.addcmd({parser: True}, parser, True)
        else:
            if want_cdp is True and have_cdp is False:
                self.addcmd({parser: want_cdp}, parser, not want_cdp)
            elif want_cdp is False and have_cdp is None:
                self.addcmd({parser: not want_cdp}, parser, not want_cdp)
            elif want_cdp is None and have_cdp is False:
                if self.state in ["overridden", "deleted"] and not want:
                    self.addcmd({parser: not have_cdp}, parser, have_cdp)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2019, Dag Wieers (@dagwieers) <dag@wieers.com>
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {"metadata_version": "1.1", "status": ["preview"], "supported_by": "community"}

DOCUMENTATION = r"""
---
module: mso_schema_site_bd
short_description: Manage site-local Bridge Domains (BDs) in schema template
description:
- Manage site-local BDs in schema template on Cisco ACI Multi-Site.
author:
- Dag Wieers (@dagwieers)
options:
  schema:
    description:
    - The name of the schema.
    type: str
    required: true
  site:
    description:
    - The name of the site.
    type: str
    required: true
  template:
    description:
    - The name of the template.
    type: str
    required: true
  bd:
    description:
    - The name of the BD to manage.
    type: str
    aliases: [ name ]
  host_route:
    description:
    - Whether host-based routing is enabled.
    type: bool
  svi_mac:
    description:
    - SVI MAC Address
    type: str
  state:
    description:
    - Use C(present) or C(absent) for adding or removing.
    - Use C(query) for listing an object or multiple objects.
    type: str
    choices: [ absent, present, query ]
    default: present
seealso:
- module: cisco.mso.mso_schema_site
- module: cisco.mso.mso_schema_site_bd_l3out
- module: cisco.mso.mso_schema_site_bd_subnet
- module: cisco.mso.mso_schema_template_bd
extends_documentation_fragment: cisco.mso.modules
"""

EXAMPLES = r"""
- name: Add a new site BD
  cisco.mso.mso_schema_site_bd:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema1
    site: Site1
    template: Template1
    bd: BD1
    state: present

- name: Remove a site BD
  cisco.mso.mso_schema_site_bd:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema1
    site: Site1
    template: Template1
    bd: BD1
    state: absent

- name: Query a specific site BD
  cisco.mso.mso_schema_site_bd:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema1
    site: Site1
    template: Template1
    bd: BD1
    state: query
  register: query_result

- name: Query all site BDs
  cisco.mso.mso_schema_site_bd:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema1
    site: Site1
    template: Template1
    state: query
  register: query_result
"""

RETURN = r"""
"""

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.mso.plugins.module_utils.mso import MSOModule, mso_argument_spec


def main():
    argument_spec = mso_argument_spec()
    argument_spec.update(
        schema=dict(type="str", required=True),
        site=dict(type="str", required=True),
        template=dict(type="str", required=True),
        bd=dict(type="str", aliases=["name"]),  # This parameter is not required for querying all objects
        host_route=dict(type="bool"),
        svi_mac=dict(type="str"),
        state=dict(type="str", default="present", choices=["absent", "present", "query"]),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_if=[
            ["state", "absent", ["bd"]],
            ["state", "present", ["bd"]],
        ],
    )

    schema = module.params.get("schema")
    site = module.params.get("site")
    template = module.params.get("template").replace(" ", "")
    bd = module.params.get("bd")
    host_route = module.params.get("host_route")
    svi_mac = module.params.get("svi_mac")
    state = module.params.get("state")

    mso = MSOModule(module)

    # Get schema objects
    schema_id, schema_path, schema_obj = mso.query_schema(schema)

    # Get template
    templates = [t.get("name") for t in schema_obj.get("templates")]
    if template not in templates:
        mso.fail_json(msg="Provided template '{0}' does not exist. Existing templates: {1}".format(template, ", ".join(templates)))

    # Get site
    site_id = mso.lookup_site(site)

    # Get site_idx
    if not schema_obj.get("sites"):
        mso.fail_json(msg="No site associated with template '{0}'. Associate the site with the template using mso_schema_site.".format(template))
    sites = [(s.get("siteId"), s.get("templateName")) for s in schema_obj.get("sites")]
    if (site_id, template) not in sites:
        mso.fail_json(msg="Provided site-template association '{0}-{1}' does not exist.".format(site, template))

    # Schema-access uses indexes
    site_idx = sites.index((site_id, template))
    # Path-based access uses site_id-template
    site_template = "{0}-{1}".format(site_id, template)

    # Get BD
    bd_ref = mso.bd_ref(schema_id=schema_id, template=template, bd=bd)
    bds = [v.get("bdRef") for v in schema_obj.get("sites")[site_idx]["bds"]]
    if bd is not None and bd_ref in bds:
        bd_idx = bds.index(bd_ref)
        bd_path = "/sites/{0}/bds/{1}".format(site_template, bd)
        mso.existing = schema_obj.get("sites")[site_idx]["bds"][bd_idx]
        mso.existing["bdRef"] = mso.dict_from_ref(mso.existing.get("bdRef"))

    if state == "query":
        if bd is None:
            mso.existing = schema_obj.get("sites")[site_idx]["bds"]
            for bd in mso.existing:
                bd["bdRef"] = mso.dict_from_ref(bd.get("bdRef"))
        elif not mso.existing:
            mso.fail_json(msg="BD '{bd}' not found".format(bd=bd))
        mso.exit_json()

    bds_path = "/sites/{0}/bds".format(site_template)
    ops = []

    mso.previous = mso.existing
    if state == "absent":
        if mso.existing:
            mso.sent = mso.existing = {}
            ops.append(dict(op="remove", path=bd_path))

    elif state == "present":
        if not mso.existing:
            if host_route is None:
                host_route = False

        payload = dict(
            bdRef=dict(
                schemaId=schema_id,
                templateName=template,
                bdName=bd,
            ),
            hostBasedRouting=host_route,
        )
        if svi_mac is not None:
            payload.update(mac=svi_mac)

        mso.sanitize(payload, collate=True)

        if mso.existing:
            ops.append(dict(op="replace", path=bd_path, value=mso.sent))
        else:
            ops.append(dict(op="add", path=bds_path + "/-", value=mso.sent))

        mso.existing = mso.proposed

    if not module.check_mode and mso.existing != mso.previous:
        mso.request(schema_path, method="PATCH", data=ops)

    mso.exit_json()


if __name__ == "__main__":
    main()



#!/usr/bin/python

# Copyright (c) 2018, Jan Christian Grünhage <jan.christian@gruenhage.xyz>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
author: "Jan Christian Grünhage (@jcgruenhage)"
module: matrix
short_description: Send notifications to matrix
description:
  - This module sends HTML formatted notifications to matrix rooms.
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  msg_plain:
    type: str
    description:
      - Plain text form of the message to send to matrix, usually markdown.
    required: true
  msg_html:
    type: str
    description:
      - HTML form of the message to send to matrix.
    required: true
  room_id:
    type: str
    description:
      - ID of the room to send the notification to.
    required: true
  hs_url:
    type: str
    description:
      - URL of the homeserver, where the CS-API is reachable.
    required: true
  token:
    type: str
    description:
      - Authentication token for the API call. If provided, O(user_id) and O(password) are not required.
  user_id:
    type: str
    description:
      - The user ID of the user.
  password:
    type: str
    description:
      - The password to log in with.
requirements:
  - matrix-client (Python library)
"""

EXAMPLES = r"""
- name: Send matrix notification with token
  community.general.matrix:
    msg_plain: "**hello world**"
    msg_html: "<b>hello world</b>"
    room_id: "!12345678:server.tld"
    hs_url: "https://matrix.org"
    token: "{{ matrix_auth_token }}"

- name: Send matrix notification with user_id and password
  community.general.matrix:
    msg_plain: "**hello world**"
    msg_html: "<b>hello world</b>"
    room_id: "!12345678:server.tld"
    hs_url: "https://matrix.org"
    user_id: "ansible_notification_bot"
    password: "{{ matrix_auth_password }}"
"""

RETURN = r"""
"""
import traceback

from ansible.module_utils.basic import AnsibleModule, missing_required_lib

MATRIX_IMP_ERR = None
try:
    from matrix_client.client import MatrixClient
except ImportError:
    MATRIX_IMP_ERR = traceback.format_exc()
    matrix_found = False
else:
    matrix_found = True


def run_module():
    module_args = dict(
        msg_plain=dict(type="str", required=True),
        msg_html=dict(type="str", required=True),
        room_id=dict(type="str", required=True),
        hs_url=dict(type="str", required=True),
        token=dict(type="str", no_log=True),
        user_id=dict(type="str"),
        password=dict(type="str", no_log=True),
    )

    result = dict(changed=False, message="")

    module = AnsibleModule(
        argument_spec=module_args,
        mutually_exclusive=[["password", "token"]],
        required_one_of=[["password", "token"]],
        required_together=[["user_id", "password"]],
        supports_check_mode=True,
    )

    if not matrix_found:
        module.fail_json(msg=missing_required_lib("matrix-client"), exception=MATRIX_IMP_ERR)

    if module.check_mode:
        return result

    # create a client object
    client = MatrixClient(module.params["hs_url"])
    if module.params["token"] is not None:
        client.api.token = module.params["token"]
    else:
        client.login(module.params["user_id"], module.params["password"], sync=False)

    # make sure we are in a given room and return a room object for it
    room = client.join_room(module.params["room_id"])
    # send an html formatted messages
    room.send_html(module.params["msg_html"], module.params["msg_plain"])

    module.exit_json(**result)


def main():
    run_module()


if __name__ == "__main__":
    main()



# (c) 2023, Alexei Znamensky <russoz@gmail.com>
# Copyright (c) 2023, Ansible Project
# Simplified BSD License (see LICENSES/BSD-2-Clause.txt or https://opensource.org/licenses/BSD-2-Clause)
# SPDX-License-Identifier: BSD-2-Clause

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import copy
import typing as t


class _Variable:
    NOTHING = object()

    def __init__(
        self,
        diff: bool = False,
        output: bool = True,
        change: bool | None = None,
        fact: bool = False,
        verbosity: int = 0,
    ):
        self.init = False
        self.initial_value: t.Any = None
        self.value: t.Any = None

        self.diff: bool = None  # type: ignore # will be changed in set_meta() call
        self._change: bool | None = None
        self.output: bool = None  # type: ignore # will be changed in set_meta() call
        self.fact: bool = None  # type: ignore # will be changed in set_meta() call
        self._verbosity: int = None  # type: ignore # will be changed in set_meta() call
        self.set_meta(output=output, diff=diff, change=change, fact=fact, verbosity=verbosity)

    def getchange(self) -> bool:
        return self.diff if self._change is None else self._change

    def setchange(self, value: bool | None) -> None:
        self._change = value

    def getverbosity(self) -> int:
        return self._verbosity

    def setverbosity(self, v: int) -> None:
        if not (0 <= v <= 4):
            raise ValueError("verbosity must be an int in the range 0 to 4")
        self._verbosity = v

    change = property(getchange, setchange)
    verbosity = property(getverbosity, setverbosity)

    def set_meta(
        self,
        output: bool | None = None,
        diff: bool | None = None,
        change: bool | None = None,
        fact: bool | None = None,
        initial_value: t.Any = NOTHING,
        verbosity: int | None = None,
    ) -> None:
        """Set the metadata for the variable

        Args:
            output (bool, optional): flag indicating whether the variable should be in the output of the module. Defaults to None.
            diff (bool, optional): flag indicating whether to generate diff mode output for this variable. Defaults to None.
            change (bool, optional): flag indicating whether to track if changes happened to this variable. Defaults to None.
            fact (bool, optional): flag indicating whether the variable should be exposed as a fact of the module. Defaults to None.
            initial_value (any, optional): initial value of the variable, to be used with `change`. Defaults to NOTHING.
            verbosity (int, optional): level of verbosity in which this variable is reported by the module as `output`, `fact` or `diff`. Defaults to None.
        """
        if output is not None:
            self.output = output
        if change is not None:
            self.change = change
        if diff is not None:
            self.diff = diff
        if fact is not None:
            self.fact = fact
        if initial_value is not _Variable.NOTHING:
            self.initial_value = copy.deepcopy(initial_value)
        if verbosity is not None:
            self.verbosity = verbosity

    def as_dict(self, meta_only: bool = False) -> dict[str, t.Any]:
        d = {
            "diff": self.diff,
            "change": self.change,
            "output": self.output,
            "fact": self.fact,
            "verbosity": self.verbosity,
        }
        if not meta_only:
            d["initial_value"] = copy.deepcopy(self.initial_value)
            d["value"] = self.value
        return d

    def set_value(self, value: t.Any) -> t.Self:
        if not self.init:
            self.initial_value = copy.deepcopy(value)
            self.init = True
        self.value = value
        return self

    def is_visible(self, verbosity: int) -> bool:
        return self.verbosity <= verbosity

    @property
    def has_changed(self) -> bool:
        return self.change and (self.initial_value != self.value)

    @property
    def diff_result(self) -> dict[str, t.Any] | None:
        if self.diff and self.has_changed:
            return {"before": self.initial_value, "after": self.value}
        return None

    def __str__(self) -> str:
        return (
            f"<Variable: value={self.value!r}, initial={self.initial_value!r}, diff={self.diff}, "
            f"output={self.output}, change={self.change}, verbosity={self.verbosity}>"
        )


class VarDict:
    reserved_names = (
        "__vars__",
        "_var",
        "var",
        "set_meta",
        "get_meta",
        "set",
        "output",
        "diff",
        "facts",
        "has_changed",
        "as_dict",
    )

    def __init__(self) -> None:
        self.__vars__: dict[str, _Variable] = dict()

    def __getitem__(self, item: str):
        return self.__vars__[item].value

    def __setitem__(self, key: str, value) -> None:
        self.set(key, value)

    def __getattr__(self, item: str):
        try:
            return self.__vars__[item].value
        except KeyError:
            return getattr(super(), item)

    def __setattr__(self, key: str, value) -> None:
        if key == "__vars__":
            super().__setattr__(key, value)
        else:
            self.set(key, value)

    def _var(self, name: str) -> _Variable:
        return self.__vars__[name]

    def var(self, name: str) -> dict[str, t.Any]:
        return self._var(name).as_dict()

    def set_meta(self, name: str, **kwargs):
        """Set the metadata for the variable

        Args:
            name (str): name of the variable having its metadata changed
            output (bool, optional): flag indicating whether the variable should be in the output of the module. Defaults to None.
            diff (bool, optional): flag indicating whether to generate diff mode output for this variable. Defaults to None.
            change (bool, optional): flag indicating whether to track if changes happened to this variable. Defaults to None.
            fact (bool, optional): flag indicating whether the variable should be exposed as a fact of the module. Defaults to None.
            initial_value (any, optional): initial value of the variable, to be used with `change`. Defaults to NOTHING.
            verbosity (int, optional): level of verbosity in which this variable is reported by the module as `output`, `fact` or `diff`. Defaults to None.
        """
        self._var(name).set_meta(**kwargs)

    def get_meta(self, name: str) -> dict[str, t.Any]:
        return self._var(name).as_dict(meta_only=True)

    def set(self, name: str, value, **kwargs) -> None:
        """Set the value and optionally metadata for a variable. The variable is not required to exist prior to calling `set`.

        For details on the accepted metada see the documentation for method `set_meta`.

        Args:
            name (str): name of the variable being changed
            value (any): the value of the variable, it can be of any type

        Raises:
            ValueError: Raised if trying to set a variable with a reserved name.
        """
        if name in self.reserved_names:
            raise ValueError(f"Name {name} is reserved")
        if name in self.__vars__:
            var = self._var(name)
            var.set_meta(**kwargs)
        else:
            var = _Variable(**kwargs)
        var.set_value(value)
        self.__vars__[name] = var

    def output(self, verbosity: int = 0) -> dict[str, t.Any]:
        return {n: v.value for n, v in self.__vars__.items() if v.output and v.is_visible(verbosity)}

    def diff(self, verbosity: int = 0) -> dict[str, dict[str, t.Any]] | None:
        diff_results = [
            (n, v.diff_result) for n, v in self.__vars__.items() if v.diff_result and v.is_visible(verbosity)
        ]
        if diff_results:
            before = {n: dr["before"] for n, dr in diff_results}
            after = {n: dr["after"] for n, dr in diff_results}
            return {"before": before, "after": after}
        return None

    def facts(self, verbosity: int = 0) -> dict[str, t.Any] | None:
        facts_result = {n: v.value for n, v in self.__vars__.items() if v.fact and v.is_visible(verbosity)}
        return facts_result if facts_result else None

    @property
    def has_changed(self) -> bool:
        return any(var.has_changed for var in self.__vars__.values())

    def as_dict(self) -> dict[str, t.Any]:
        return {name: var.value for name, var in self.__vars__.items()}



# -*- coding: utf-8 -*-

# Copyright: (c) 2024, Sabari Jaganathan (@sajagana) <sajagana@cisco.com>
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type
import copy


def generate_api_endpoint(path, **kwargs):
    """
    Generates an API endpoint with query strings based on the provided keyword arguments.

    :param path: The base URL of the API endpoint. -> Str
    :param kwargs: Keyword arguments representing query parameters. -> Dict
    :return: A string representing the full API endpoint with query parameters. -> Str
    """
    return path if not kwargs else "{0}?{1}".format(path, "&".join(["{0}={1}".format(key, value) for key, value in kwargs.items()]))


def append_update_ops_data(ops, existing_data, update_path, replace_data=None, remove_data=None):
    """
    Append Update ops payload data.
    :param ops: Variable which contains the PATCH replace actions for the update operation -> List
    :param existing_data: Variable which contains the existing data -> Dict
    :param update_path: The object path is used to update an existing object -> Str
    :param replace_data: Defaults to None when not specified, expected a dictionary object. Which contains the attribute to be updated and its new value -> Dict
    :param remove_data: Defaults to None when not specified, expected a list of string or tuple, value used to clear the existing configuration -> List
    :return: None
                If attributes is not empty then the ops and existing_data are updated with the input value.

    Sample Existing Data:
    ---------------------
    existing_data = {
        "name": "name",
        "description": "description",
        "bfdMultiHopPol": {
            "adminState": "enabled",
            "minRxInterval": 250,
            "ifControl": {"adminState": "enabled"},
        },
        "bfdPol": {
            "adminState": "enabled",
            "detectionMultiplier": 3,
        }
    }

    ops = []
    update_path = "/tenantPolicyTemplate/template/l3OutIntfPolGroups/0"
    replace_data = {
        ("name"): "new_name",
        "description": "new_description",
        ("ospfIntfPol"): dict(ifControl=dict(adminState="disabled"), cost=0),
    }
    remove_data = [("bfdMultiHopPol", "ifControl", "adminState"), "bfdPol"]

    append_update_ops_data(ops, existing_data, update_path, replace_data, remove_data)

    Standard Output Data:
    ---------------------
    {
        "bfdMultiHopPol": {
            "adminState": "enabled",
            "minRxInterval": 250,
            "ifControl": {},
        },
        "name": "new_name",
        "description": "new_description",
        "ospfIntfPol": {
            "ifControl": {
                "adminState": "disabled",
            },
            "cost": 0,
        },
    }

    API Input Data:
    ---------------
    [
        {"op": "replace", "path": "/tenantPolicyTemplate/template/l3OutIntfPolGroups/1/name", "value": "new_name"},
        {"op": "replace", "path": "/tenantPolicyTemplate/template/l3OutIntfPolGroups/1/description", "value": "new_description"},
        {
            "op": "replace",
            "path": "/tenantPolicyTemplate/template/l3OutIntfPolGroups/1/ospfIntfPol",
            "value": {"ifControl": {"adminState": "disabled"}, "cost": 0},
        },
        {"op": "remove", "path": "/tenantPolicyTemplate/template/l3OutIntfPolGroups/1/bfdPol"},
        {"op": "remove", "path": "/tenantPolicyTemplate/template/l3OutIntfPolGroups/1/bfdMultiHopPol/ifControl/adminState"},
    ]
    """

    def recursive_add_replace(data, path, keys, new_value):
        key = keys[0]
        if len(keys) == 1:
            # Update the existing configuration
            if new_value is not None and data.get(key) != new_value:
                operation = "replace" if data.get(key) is not None else "add"
                data[key] = new_value
                ops.append(
                    dict(
                        op=operation,
                        path="{}/{}".format(path, key),
                        value=copy.deepcopy(new_value),
                    )
                )
        elif key in data:
            recursive_add_replace(data[key], "{}/{}".format(path, key), keys[1:], new_value)

    def recursive_delete(data, path, keys):
        key = keys[0]
        if len(keys) == 1:
            # Clear the existing configuration
            if key in data:
                data.pop(key)
                ops.append(
                    dict(
                        op="remove",
                        path="{}/{}".format(path, key),
                    )
                )
        elif key in data:
            recursive_delete(data[key], "{}/{}".format(path, key), keys[1:])

    if replace_data:
        if not isinstance(replace_data, dict):
            raise TypeError("replace_data must be a dict")

        for key, value in replace_data.items():
            recursive_add_replace(existing_data, update_path, key if isinstance(key, tuple) else (key,), value)

    if remove_data:
        if not isinstance(remove_data, list):
            raise TypeError("remove_data must be a list of string or tuples")

        for key in remove_data:
            recursive_delete(existing_data, update_path, key if isinstance(key, tuple) else (key,))


def check_if_all_elements_are_none(values):
    """
    Checks if all elements in the provided iterable are None

    :param values: An iterable containing values to be checked -> Iterable[Any]
    :return: True if all elements are None, False otherwise -> Bool
    """
    return all(value is None for value in values)


def snake_to_camel(snake_str, upper_case_components=None):
    if snake_str is not None and "_" in snake_str:
        if upper_case_components is None:
            upper_case_components = []
        components = snake_str.split("_")
        camel_case_str = components[0]

        for component in components[1:]:
            if component in upper_case_components:
                camel_case_str += component.upper()
            else:
                camel_case_str += component.title()

        return camel_case_str
    else:
        return snake_str


def delete_none_values(obj_to_sanitize, recursive=True):
    """
    Removes keys with None values from a Python object, which can be either a list or a dictionary.
    Optionally performs the operation recursively on nested structures.

    :param obj_to_sanitize: The Python object to sanitize from None values. -> List or Dict
    :param recursive: A boolean flag indicating whether to recursively sanitize nested objects. Defaults to True. -> bool
    :return: A sanitized copy of the original Python object, with all keys with None values removed. -> List or Dict
    """
    if isinstance(obj_to_sanitize, dict):
        sanitized_dict = {}
        for item_key, item_value in obj_to_sanitize.items():
            if recursive and isinstance(item_value, (dict, list)):
                sanitized_dict[item_key] = delete_none_values(item_value, recursive)
            elif item_value is not None:
                sanitized_dict[item_key] = item_value
        return sanitized_dict

    elif isinstance(obj_to_sanitize, list):
        sanitized_list = []
        for item in obj_to_sanitize:
            if recursive and isinstance(item, (dict, list)):
                sanitized_list.append(delete_none_values(item, recursive))
            elif item is not None:
                sanitized_list.append(item)
        return sanitized_list

    else:
        raise TypeError("Object to sanitize must be of type list or dict. Got {}".format(type(obj_to_sanitize)))


def format_annotations_list(mso_values, annotations):
    """
    Format the list of annotations for the payload to be sent
    :param mso_values: The payload to be sent -> Dict
    :param annotations: The initial list of annotations -> Dict
    :return: Assign the formated list of annotations to the payload under key 'tagAnnotations' -> None
    """
    if annotations:
        mso_values["tagAnnotations"] = [
            {
                "tagKey": annotation.get("key"),
                "tagValue": annotation.get("value"),
            }
            for annotation in annotations
        ]



#!/usr/bin/python
#
# Copyright (c) 2014-2015, Epic Games, Inc.
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: circonus_annotation
short_description: Create an annotation in Circonus
description:
  - Create an annotation event with a given category, title and description. Optionally start, end or durations can be provided.
author: "Nick Harring (@NickatEpic)"
requirements:
  - requests >= 2.0.0
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: none
  diff_mode:
    support: none
options:
  api_key:
    type: str
    description:
      - Circonus API key.
    required: true
  category:
    type: str
    description:
      - Annotation Category.
    required: true
  description:
    type: str
    description:
      - Description of annotation.
    required: true
  title:
    type: str
    description:
      - Title of annotation.
    required: true
  start:
    type: int
    description:
      - Unix timestamp of event start.
      - If not specified, it defaults to "now".
  stop:
    type: int
    description:
      - Unix timestamp of event end.
      - If not specified, it defaults to "now" + O(duration).
  duration:
    type: int
    description:
      - Duration in seconds of annotation.
    default: 0
"""
EXAMPLES = r"""
- name: Create a simple annotation event with a source, defaults to start and end time of now
  community.general.circonus_annotation:
    api_key: XXXXXXXXXXXXXXXXX
    title: App Config Change
    description: This is a detailed description of the config change
    category: This category groups like annotations

- name: Create an annotation with a duration of 5 minutes and a default start time of now
  community.general.circonus_annotation:
    api_key: XXXXXXXXXXXXXXXXX
    title: App Config Change
    description: This is a detailed description of the config change
    category: This category groups like annotations
    duration: 300

- name: Create an annotation with a start_time and end_time
  community.general.circonus_annotation:
    api_key: XXXXXXXXXXXXXXXXX
    title: App Config Change
    description: This is a detailed description of the config change
    category: This category groups like annotations
    start_time: 1395940006
    end_time: 1395954407
"""

RETURN = r"""
annotation:
  description: Details about the created annotation.
  returned: success
  type: complex
  contains:
    _cid:
      description: Annotation identifier.
      returned: success
      type: str
      sample: /annotation/100000
    _created:
      description: Creation timestamp.
      returned: success
      type: int
      sample: 1502236928
    _last_modified:
      description: Last modification timestamp.
      returned: success
      type: int
      sample: 1502236928
    _last_modified_by:
      description: Last modified by.
      returned: success
      type: str
      sample: /user/1000
    category:
      description: Category of the created annotation.
      returned: success
      type: str
      sample: alerts
    title:
      description: Title of the created annotation.
      returned: success
      type: str
      sample: WARNING
    description:
      description: Description of the created annotation.
      returned: success
      type: str
      sample: Host is down.
    start:
      description: Timestamp, since annotation applies.
      returned: success
      type: int
      sample: Host is down.
    stop:
      description: Timestamp, since annotation ends.
      returned: success
      type: str
      sample: Host is down.
    rel_metrics:
      description: Array of metrics related to this annotation, each metrics is a string.
      returned: success
      type: list
      sample:
        - 54321_kbps
"""

import json
import time
import traceback

from ansible_collections.community.general.plugins.module_utils._version import LooseVersion

REQUESTS_IMP_ERR = None
try:
    import requests

    HAS_REQUESTS = True
except ImportError:
    REQUESTS_IMP_ERR = traceback.format_exc()
    HAS_REQUESTS = False

from ansible.module_utils.basic import AnsibleModule, missing_required_lib


def check_requests_dep(module):
    """Check if an adequate requests version is available"""
    if not HAS_REQUESTS:
        module.fail_json(msg=missing_required_lib("requests"), exception=REQUESTS_IMP_ERR)
    else:
        required_version = "2.0.0"
        if LooseVersion(requests.__version__) < LooseVersion(required_version):
            module.fail_json(
                msg=f"'requests' library version should be >= {required_version}, found: {requests.__version__}."
            )


def post_annotation(annotation, api_key):
    """Takes annotation dict and api_key string"""
    base_url = "https://api.circonus.com/v2"
    anootate_post_endpoint = "/annotation"
    resp = requests.post(base_url + anootate_post_endpoint, headers=build_headers(api_key), data=json.dumps(annotation))
    resp.raise_for_status()
    return resp


def create_annotation(module):
    """Takes ansible module object"""
    annotation = {}
    duration = module.params["duration"]
    if module.params["start"] is not None:
        start = module.params["start"]
    else:
        start = int(time.time())
    if module.params["stop"] is not None:
        stop = module.params["stop"]
    else:
        stop = int(time.time()) + duration
    annotation["start"] = start
    annotation["stop"] = stop
    annotation["category"] = module.params["category"]
    annotation["description"] = module.params["description"]
    annotation["title"] = module.params["title"]
    return annotation


def build_headers(api_token):
    """Takes api token, returns headers with it included."""
    headers = {
        "X-Circonus-App-Name": "ansible",
        "Host": "api.circonus.com",
        "X-Circonus-Auth-Token": api_token,
        "Accept": "application/json",
    }
    return headers


def main():
    """Main function, dispatches logic"""
    module = AnsibleModule(
        argument_spec=dict(
            start=dict(type="int"),
            stop=dict(type="int"),
            category=dict(required=True),
            title=dict(required=True),
            description=dict(required=True),
            duration=dict(default=0, type="int"),
            api_key=dict(required=True, no_log=True),
        )
    )

    check_requests_dep(module)

    annotation = create_annotation(module)
    try:
        resp = post_annotation(annotation, module.params["api_key"])
    except requests.exceptions.RequestException as e:
        module.fail_json(msg="Request Failed", reason=f"{e}", exception=traceback.format_exc())
    module.exit_json(changed=True, annotation=resp.json())


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

# Copyright: Contributors to the Ansible project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

# try:
#     import botocore
# except ImportError:
#     pass  # Modules are responsible for handling this.

from ..botocore import is_boto3_error_code
from ..errors import AWSErrorHandler
from ..exceptions import AnsibleAWSError


class AnsibleAutoScalingError(AnsibleAWSError):
    pass


class AutoScalingErrorHandler(AWSErrorHandler):
    _CUSTOM_EXCEPTION = AnsibleAutoScalingError

    @classmethod
    def _is_missing(cls):
        return is_boto3_error_code("NoSuchEntity")



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2012, Michael DeHaan <michael.dehaan@gmail.com>, and others
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r'''
---
module: win_ping
short_description: A windows version of the classic ping module
description:
  - Checks management connectivity of a windows host.
  - This is NOT ICMP ping, this is just a trivial test module.
  - Since C(ansible.windows 3.6.0) this module can also run on non-Windows
    targets with PowerShell 7 present. Targeting non-Windows hosts requires
    Ansible C(2.22.0) or newer. Otherwise use the M(ansible.builtin.ping)
    module instead.
options:
  data:
    description:
      - Alternate data to return instead of 'pong'.
      - If this parameter is set to C(crash), the module will cause an exception.
    type: str
    default: pong
seealso:
- module: ansible.builtin.ping
author:
- Chris Church (@cchurch)
'''

EXAMPLES = r'''
# Test connectivity to a windows host
# ansible winserver -m ansible.windows.win_ping

- name: Example from an Ansible Playbook
  ansible.windows.win_ping:

- name: Induce an exception to see what happens
  ansible.windows.win_ping:
    data: crash
'''

RETURN = r'''
ping:
    description: Value provided with the data parameter.
    returned: success
    type: str
    sample: pong
'''



#
# (c) 2018 Red Hat Inc.
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import absolute_import, division, print_function


__metaclass__ = type

import os
import re
import time

from ansible.errors import AnsibleError
from ansible.module_utils.common.text.converters import to_text
from ansible.module_utils.six.moves.urllib.parse import urlsplit
from ansible.plugins.action.normal import ActionModule as _ActionModule
from ansible.utils.display import Display
from ansible.utils.hashing import checksum, checksum_s


# Ansible 2.19+ requires trust_as_template for Jinja2 processing
try:
    from ansible.template import trust_as_template

    HAS_TRUST_AS_TEMPLATE = True
except ImportError:
    HAS_TRUST_AS_TEMPLATE = False

display = Display()


def _netcommon_remove_internal_keys_fallback(data):
    """Mirror ansible.vars.clean.remove_internal_keys when that API is unavailable."""
    from ansible import constants as _ansible_constants

    # ansible-core may drop INTERNAL_RESULT_KEYS from constants; match historical defaults.
    _internal_result_keys = getattr(
        _ansible_constants, "INTERNAL_RESULT_KEYS", ("add_host", "add_group")
    )

    for key in list(data.keys()):
        if (
            key.startswith("_ansible_") and key != "_ansible_parsed"
        ) or key in _internal_result_keys:
            display.warning(
                "Removed unexpected internal key in module return: %s = %s" % (key, data[key])
            )
            del data[key]

    for key in ("warnings", "deprecations"):
        if key in data and not data[key]:
            del data[key]

    ansible_facts = data.get("ansible_facts")
    if ansible_facts:
        for key in list(ansible_facts.keys()):
            if key.startswith("discovered_interpreter_") or key.startswith(
                "ansible_discovered_interpreter_"
            ):
                del ansible_facts[key]


# ansible-core devel may relocate remove_internal_keys; keep a local fallback.
try:
    from ansible.vars.clean import remove_internal_keys as _remove_internal_keys
except ImportError:
    _remove_internal_keys = _netcommon_remove_internal_keys_fallback

DEXEC_PREFIX = "ANSIBLE_NETWORK_IMPORT_MODULES:"


class ActionModule(_ActionModule):
    def run(self, tmp=None, task_vars=None):
        config_module = hasattr(self, "_config_module") and self._config_module
        if config_module and self._task.args.get("src"):
            try:
                self._handle_src_option()
            except AnsibleError as exc:
                return dict(failed=True, msg=to_text(exc))

        host = task_vars["ansible_host"]
        dexec_eligible = self._check_dexec_eligibility(host)

        # attempt to run using dexec
        if dexec_eligible:
            # find and load the module
            filename, module = self._find_load_module()
            display.vvvv(
                "{prefix} found {action} at {fname}".format(
                    prefix=DEXEC_PREFIX,
                    action=self._task.action,
                    fname=filename,
                ),
                host,
            )
            # not using AnsibleModule, return to normal run (eg eos_bgp)
            if getattr(module, "AnsibleModule", None):
                # patch and update the module
                self._patch_update_module(module, task_vars, host)
                display.vvvv(
                    "{prefix} running {module}".format(
                        prefix=DEXEC_PREFIX, module=self._task.action
                    ),
                    host,
                )
                # execute the module, collect result
                result = self._exec_module(module)
                display.vvvv("{prefix} complete".format(prefix=DEXEC_PREFIX), host)
                display.vvvvv(
                    "{prefix} Result: {result}".format(prefix=DEXEC_PREFIX, result=result),
                    host,
                )

            else:
                dexec_eligible = False
                display.vvvv(
                    "{prefix} {module} doesn't support direct execution, disabled".format(
                        prefix=DEXEC_PREFIX, module=self._task.action
                    ),
                    host,
                )

        if not dexec_eligible:
            result = super(ActionModule, self).run(task_vars=task_vars)

        if config_module and self._task.args.get("backup") and not result.get("failed"):
            self._handle_backup_option(
                result,
                task_vars,
                self._task.args.get("backup_options"),
            )

        return result

    def _handle_backup_option(self, result, task_vars, backup_options):
        filename = None
        backup_path = None
        try:
            non_config_regexes = self._connection.cliconf.get_option("non_config_lines", task_vars)
        except (AttributeError, KeyError):
            non_config_regexes = []
        try:
            content = self._sanitize_contents(
                contents=result.pop("__backup__"), filters=non_config_regexes
            )
        except KeyError:
            raise AnsibleError("Failed while reading configuration backup")

        if backup_options:
            filename = backup_options.get("filename")
            backup_path = backup_options.get("dir_path")

        tstamp = time.strftime("%Y-%m-%d@%H:%M:%S", time.localtime(time.time()))
        if not backup_path:
            cwd = self._get_working_path()
            backup_path = os.path.join(cwd, "backup")
        if not filename:
            filename = "%s_config.%s" % (
                task_vars["inventory_hostname"],
                tstamp,
            )

        dest = os.path.join(backup_path, filename)
        if not os.path.exists(backup_path):
            os.makedirs(backup_path)

        changed = False
        # Do not overwrite the destination if the contents match.
        if not os.path.exists(dest) or checksum(dest) != checksum_s(content):
            try:
                with open(dest, "w") as output_file:
                    output_file.write(content)
            except Exception as exc:
                result["failed"] = True
                result["msg"] = "Could not write to destination file %s: %s" % (
                    dest,
                    to_text(exc),
                )
                return
            changed = True

        result["backup_path"] = dest
        result["changed"] = changed

        result["date"], result["time"] = tstamp.split("@")
        if not (backup_options and backup_options.get("filename")):
            result["filename"] = os.path.basename(result["backup_path"])
            result["shortname"] = os.path.splitext(result["backup_path"])[0]

    def _get_working_path(self):
        cwd = self._loader.get_basedir()
        if self._task._role is not None:
            cwd = self._task._role._role_path
        return cwd

    def _handle_src_option(self, convert_data=True):
        src = self._task.args.get("src")
        working_path = self._get_working_path()

        if os.path.isabs(src) or urlsplit(src).scheme:
            source = src
        else:
            source = self._loader.path_dwim_relative(working_path, "templates", src)
            if not source:
                source = self._loader.path_dwim_relative(working_path, src)

        if not os.path.exists(source):
            raise AnsibleError("path specified in src not found")

        try:
            with open(source, "r") as f:
                template_data = to_text(f.read())
        except IOError as e:
            raise AnsibleError(
                "unable to load src file {0}, I/O error({1}): {2}".format(
                    source, e.errno, e.strerror
                )
            )

        # Create a template search path in the following order:
        # [working_path, self_role_path, dependent_role_paths, dirname(source)]
        searchpath = [working_path]
        if self._task._role is not None:
            searchpath.append(self._task._role._role_path)
            if hasattr(self._task, "_block:"):
                dep_chain = self._task._block.get_dep_chain()
                if dep_chain is not None:
                    for role in dep_chain:
                        searchpath.append(role._role_path)
        searchpath.append(os.path.dirname(source))
        self._templar.environment.loader.searchpath = searchpath
        # help_text="Use `ansible.builtin.template` instead.", supported after 2.18 update as per support
        display.deprecated(
            msg="Direct processing of templates via `src` is deprecated, use `ansible.builtin.template` instead.",
            date="2028-01-01",
            collection_name="ansible.netcommon",
        )
        # Ansible 2.19+ requires marking template data as trusted for Jinja2 processing
        # In earlier versions, template data is processed directly
        if HAS_TRUST_AS_TEMPLATE:
            template_data = trust_as_template(template_data)

        self._task.args["src"] = self._templar.template(template_data)

    def _get_network_os(self, task_vars):
        if "network_os" in self._task.args and self._task.args["network_os"]:
            display.vvvv("Getting network OS from task argument")
            network_os = self._task.args["network_os"]
        elif self._play_context.network_os:
            display.vvvv("Getting network OS from inventory")
            network_os = self._play_context.network_os
        elif (
            "network_os" in task_vars.get("ansible_facts", {})
            and task_vars["ansible_facts"]["network_os"]
        ):
            display.vvvv("Getting network OS from fact")
            network_os = task_vars["ansible_facts"]["network_os"]
        else:
            raise AnsibleError("ansible_network_os must be specified on this host")

        return network_os

    def _check_dexec_eligibility(self, host):
        """Check if current python and task are eligble"""
        dexec = self.get_connection_option("import_modules")

        # log early about dexec
        if dexec:
            display.vvvv("{prefix} enabled".format(prefix=DEXEC_PREFIX), host)

            # disable dexec when running async
            if self._task.async_val:
                dexec = False
                display.vvvv(
                    "{prefix} disabled for a task using async".format(prefix=DEXEC_PREFIX),
                    host=host,
                )
        else:
            display.vvvv("{prefix} disabled".format(prefix=DEXEC_PREFIX), host)
            display.vvvv(
                "{prefix} module execution time may be extended".format(prefix=DEXEC_PREFIX),
                host,
            )

        return dexec

    def _find_load_module(self):
        """Use the task action to find a module
        and import it.

        :return filename: The module's filename
        :rtype filename: str
        :return module: The loaded module file
        :rtype module: module
        """
        import importlib

        mloadr = self._shared_loader_obj.module_loader

        # 2.10
        try:
            context = mloadr.find_plugin_with_context(
                self._task.action, collection_list=self._task.collections
            )
            filename = context.plugin_resolved_path
            module = importlib.import_module(context.plugin_resolved_name)
        # 2.9
        except AttributeError:
            fullname, filename = mloadr.find_plugin_with_name(
                self._task.action, collection_list=self._task.collections
            )
            module = importlib.import_module(fullname)
        return filename, module

    def _patch_update_module(self, module, task_vars, host):
        """Update a module instance, replacing it's AnsibleModule
        with one that doesn't load params

        :param module: An loaded module
        :type module: A module file that was loaded
        :param task_vars: The vars provided to the task
        :type task_vars: dict
        """
        import copy

        from ansible.module_utils.basic import AnsibleModule as _AnsibleModule

        # build an AnsibleModule that doesn't load params
        class DirectExecutionModule(_AnsibleModule):
            def _load_params(self):
                """Don't load params for action plugin use case - params set externally"""
                display.vvvv(
                    "{prefix} _load_params skipped for action plugin in direct execution".format(
                        prefix=DEXEC_PREFIX
                    ),
                    host,
                )
                from ansible.module_utils import basic as _basic

                if getattr(_basic, "_PARSED_MODULE_ARGS", None) is None:
                    _basic._PARSED_MODULE_ARGS = {}

            def _record_module_result(self, o):
                """Override new 2.19.1+ hook to directly record the module result as a module attr."""
                module._raw_result = o

        # update the task args w/ all the magic vars
        self._update_module_args(self._task.action, self._task.args, task_vars)

        # set the params of the ansible module cause we're not using stdin
        # use a copy so the module doesn't change the original task args
        DirectExecutionModule.params = copy.deepcopy(self._task.args)

        # give the module our revised AnsibleModule
        module.AnsibleModule = DirectExecutionModule

    def _exec_module(self, module):
        """exec the module's main() since modules
        print their result, we need to replace stdout
        with a buffer. If main() fails, we assume that as stderr
        Once we collect stdout/stderr, use our super to json load
        it or handle a traceback

        :param module: An loaded module
        :type module: A module file that was loaded
        :return module_result: The result of the module
        :rtype module_result: dict
        """
        import io
        import sys

        # preserve stdout/stderr to swap and restore later, create private buffers to capture
        sys_stdout = sys.stdout
        sys_stderr = sys.stderr
        captured_stdout = io.StringIO()
        captured_stderr = io.StringIO()

        try:
            # temporarily swap stdout/stderr with private buffers
            sys.stdout = captured_stdout
            sys.stderr = captured_stderr

            module.main()  # allow unhandled module exceptions to fly; handled by TE to preserve as much error detail as possible
        except SystemExit:
            pass  # module exited cleanly
        finally:
            # restore stdout/stderr
            sys.stdout = sys_stdout
            sys.stderr = sys_stderr

        try:
            # if 2.19.1+, anything using fail_json/exit_json from the patched module should have recorded _raw_result
            data = module._raw_result
        except AttributeError:
            # if _raw_result is not available, results should be on stdout/stderr
            stdout = captured_stdout.getvalue()
            stderr = captured_stderr.getvalue()

            dict_out = {
                "stdout": stdout,
                "stdout_lines": stdout.splitlines(),
                "stderr": stderr,
                "stderr_lines": stderr.splitlines(),
            }

            # parse the response
            data = self._parse_returned_data(dict_out)

        # Clean up the response like action _execute_module
        _remove_internal_keys(data)

        # split stdout/stderr into lines if needed
        if "stdout" in data and "stdout_lines" not in data:
            # if the value is 'False', a default won't catch it.
            txt = data.get("stdout", None) or ""
            data["stdout_lines"] = txt.splitlines()
        if "stderr" in data and "stderr_lines" not in data:
            # if the value is 'False', a default won't catch it.
            txt = data.get("stderr", None) or ""
            data["stderr_lines"] = txt.splitlines()

        return data

    def _sanitize_contents(self, contents, filters):
        """remove lines from contents that match
        regexes specified in the `filters` list
        """
        for x in filters:
            contents = re.sub(x, "", contents)
        return contents.strip()



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Ansible module to manage CheckPoint Firewall (c) 2019
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import (absolute_import, division, print_function)

__metaclass__ = type

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}

DOCUMENTATION = """
---
module: cp_mgmt_export_management
short_description: Export the primary Security Management Server database or the primary Multi-Domain Server database or the single Domain database and the
                   applicable Check Point configuration.
description:
  - Export the primary Security Management Server database or the primary Multi-Domain Server database or the single Domain database and the applicable
    Check Point configuration.
  - All operations are performed over Web Services API.
  - Available from R81.10 management version.
version_added: "5.0.0"
author: "Eden Brillant (@chkp-edenbr)"
options:
  file_path:
    description:
      - Path in which the exported database file is saved.<br><font color="red">Required only</font> when not using pre-export-verification-only flag.
    type: str
  domain_name:
    description:
      - Domain name to be exported.<br><font color="red">Required only for</font> exporting a Domain from the Multi-Domain Server or backing up Domain.
    type: str
  target_version:
    description:
      - Target version.
    type: str
  include_logs:
    description:
      - Export logs without log indexes.
    type: bool
  include_logs_indexes:
    description:
      - Export logs with log indexes.
    type: bool
  include_endpoint_configuration:
    description:
      - Include export of the Endpoint Security Management configuration files.
    type: bool
  include_endpoint_database:
    description:
      - Include export of the Endpoint Security Management database.
    type: bool
  is_domain_backup:
    description:
      - If true, the exported Domain will be suitable for import on the same Multi-Domain Server only.
    type: bool
  is_smc_to_mds:
    description:
      - If true, the exported Security Management Server will be suitable for import on the Multi-Domain Server only.
    type: bool
  pre_export_verification_only:
    description:
      - If true, only runs the pre-export verifications instead of the full export.
    type: bool
  ignore_warnings:
    description:
      - Ignoring the verification warnings. By Setting this parameter to 'true' export will not be blocked by warnings.
    type: bool
extends_documentation_fragment: check_point.mgmt.checkpoint_commands
"""

EXAMPLES = """
- name: export-management
  cp_mgmt_export_management:
    domain_name: domain1
    file_path: /var/log/domain1_backup.tgz
    is_domain_backup: true
"""

RETURN = """
cp_mgmt_export_management:
  description: The checkpoint export-management output.
  returned: always.
  type: dict
"""

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.check_point.mgmt.plugins.module_utils.checkpoint import checkpoint_argument_spec_for_commands, api_command


def main():
    argument_spec = dict(
        file_path=dict(type='str'),
        domain_name=dict(type='str',),
        target_version=dict(type='str'),
        include_logs=dict(type='bool'),
        include_logs_indexes=dict(type='bool'),
        include_endpoint_configuration=dict(type='bool'),
        include_endpoint_database=dict(type='bool'),
        is_domain_backup=dict(type='bool'),
        is_smc_to_mds=dict(type='bool'),
        pre_export_verification_only=dict(type='bool'),
        ignore_warnings=dict(type='bool')
    )
    argument_spec.update(checkpoint_argument_spec_for_commands)

    module = AnsibleModule(argument_spec=argument_spec)

    command = "export-management"

    result = api_command(module, command)
    module.exit_json(**result)


if __name__ == '__main__':
    main()



# Copyright (c) 2024, Alexei Znamensky <russoz@gmail.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import typing as t
from functools import wraps

from ansible.module_utils.common.collections import is_sequence

if t.TYPE_CHECKING:
    from collections.abc import Callable, Mapping, Sequence

    ArgFormatType = Callable[[t.Any], Sequence[t.Any]]
    _T = t.TypeVar("_T")


def _ensure_list(value: _T | Sequence[_T]) -> list[_T]:
    return list(value) if is_sequence(value) else [value]  # type: ignore # TODO need type assertion for is_sequence


class _ArgFormat:
    def __init__(
        self,
        func: ArgFormatType,
        ignore_none: bool | None = True,
        ignore_missing_value: bool = False,
    ) -> None:
        self.func = func
        self.ignore_none = ignore_none
        self.ignore_missing_value = ignore_missing_value

    def __call__(self, value: t.Any | None) -> list[str]:
        ignore_none = self.ignore_none if self.ignore_none is not None else True
        if value is None and ignore_none:
            return []
        f = self.func
        return [str(x) for x in f(value)]

    def __str__(self) -> str:
        return f"<ArgFormat: func={self.func}, ignore_none={self.ignore_none}, ignore_missing_value={self.ignore_missing_value}>"

    def __repr__(self) -> str:
        return str(self)


def as_bool(
    args_true: Sequence[t.Any] | t.Any,
    args_false: Sequence[t.Any] | t.Any | None = None,
    ignore_none: bool | None = None,
) -> _ArgFormat:
    if args_false is not None:
        if ignore_none is None:
            ignore_none = False
    else:
        args_false = []
    return _ArgFormat(
        lambda value: _ensure_list(args_true) if value else _ensure_list(args_false), ignore_none=ignore_none
    )


def as_bool_not(args: Sequence[t.Any] | t.Any) -> _ArgFormat:
    return as_bool([], args, ignore_none=False)


def as_optval(arg, ignore_none: bool | None = None) -> _ArgFormat:
    return _ArgFormat(lambda value: [f"{arg}{value}"], ignore_none=ignore_none)


def as_opt_val(arg: str, ignore_none: bool | None = None) -> _ArgFormat:
    return _ArgFormat(lambda value: [arg, value], ignore_none=ignore_none)


def as_opt_eq_val(arg: str, ignore_none: bool | None = None) -> _ArgFormat:
    return _ArgFormat(lambda value: [f"{arg}={value}"], ignore_none=ignore_none)


def as_list(ignore_none: bool | None = None, min_len: int = 0, max_len: int | None = None) -> _ArgFormat:
    def func(value: t.Any) -> list[t.Any]:
        value = _ensure_list(value)
        if len(value) < min_len:
            raise ValueError(f"Parameter must have at least {min_len} element(s)")
        if max_len is not None and len(value) > max_len:
            raise ValueError(f"Parameter must have at most {max_len} element(s)")
        return value

    return _ArgFormat(func, ignore_none=ignore_none)


def as_fixed(*args: t.Any) -> _ArgFormat:
    if len(args) == 1 and is_sequence(args[0]):
        args = args[0]
    return _ArgFormat(lambda value: _ensure_list(args), ignore_none=False, ignore_missing_value=True)


def as_func(func: ArgFormatType, ignore_none: bool | None = None) -> _ArgFormat:
    return _ArgFormat(func, ignore_none=ignore_none)


def as_map(
    _map: Mapping[t.Any, Sequence[t.Any] | t.Any],
    default: Sequence[t.Any] | t.Any | None = None,
    ignore_none: bool | None = None,
) -> _ArgFormat:
    if default is None:
        default = []
    return _ArgFormat(lambda value: _ensure_list(_map.get(value, default)), ignore_none=ignore_none)


def unpack_args(func):
    @wraps(func)
    def wrapper(v):
        return func(*v)

    return wrapper


def unpack_kwargs(func):
    @wraps(func)
    def wrapper(v):
        return func(**v)

    return wrapper


def stack(fmt):
    @wraps(fmt)
    def wrapper(*args, **kwargs):
        new_func = fmt(ignore_none=True, *args, **kwargs)

        def stacking(value):
            stack = [new_func(v) for v in value if v]
            stack = [x for args in stack for x in args]
            return stack

        return _ArgFormat(stacking, ignore_none=True)

    return wrapper


def is_argformat(fmt: object) -> t.TypeGuard[_ArgFormat]:
    return isinstance(fmt, _ArgFormat)
